#include "ClientMgr.h"
#include "Protocol.h"
#include <qstringlist.h>
#include <qcombobox.h>
#include <qpushbutton.h>
#include <qsettings.h>
#include <qwidget.h>
#include <qgroupbox.h>
#include <qlayout.h>
#include <qlabel.h>
#include <qapplication.h>
#include <qsocketdevice.h>
#include <qtimer.h>
#include <qtextstream.h>
#include <qregexp.h>
#include <qlcdnumber.h>
#include <qlineedit.h>
#include <qcheckbox.h>
#include <qtooltip.h>
#include <qpen.h>
#include <qvbox.h>
#include <qscrollview.h>
#include <qsizepolicy.h>
#include <qpainter.h>
#include <qevent.h>
#include <qradiobutton.h>
#include <qbuttongroup.h>
#include <qscrollview.h>
#include <qpixmap.h>
#include <qmenubar.h>
#include <qpopupmenu.h>
#include <qmessagebox.h>
#include <qvalidator.h>
#include <string.h> // mempcy
#include <iostream>
#include <algorithm>
#include <list>
#include <sstream>
#include "ComediDevice.h"
#include "SocketMgr.h"
#include "qwt_plot.h"
#include "qwt_plot_curve.h"
#include "qwt_plot_marker.h"
#include "scrollzoomer.h"
#include "qwt_picker.h"
#include "qwt_plot_grid.h"
#include "qwt_scale_div.h"
// images.. all are char **filename_xpm[]..
#include "images/comediclient32x32.xpm"

const int ClientMgr::defInputPeriodHz = 100; // do input chan at 100Hz
const int ClientMgr::dioInputPeriodHz = 5; // do dio input chan at 10Hz
const int ClientMgr::defGraphTimeSecs = 5;
const int ClientMgr::minInputPeriodHz = 1;
const int ClientMgr::maxInputPeriodHz = 200;
const int ClientMgr::minGraphTimeSecs = 1;
const int ClientMgr::maxGraphTimeSecs = 60;

ClientMgr::ClientMgr(ClientUI *theForm_parent, const char *name)
  : QObject(theForm_parent, name), form(theForm_parent), settings(QSettings::Native), devices(NULL), numDevices(0), comm(0), data(NULL), disconnecting(false), connecting(true), graphPaused(false), dioSV(NULL), dioMgr(NULL), scrollView(NULL), serverVersion(1)
{
  settingsDialog = 0;
  form->setIcon(QPixmap(const_cast<const char **>(comediclient32x32_xpm)));
  form->setCaption("Comedi Client - v1.1");

  QGridLayout *l = new QGridLayout(form->graphFrame);  
  graph = new QwtPlot("Input Graph", form->graphFrame);
  
  graph->setCanvasBackground(QColor(29, 100, 141)); // nice blue
  form->dioControlMode->hide(); // we don't use the DIO control mode feature at this time..
  //  Insert zero line at y = 0
  QwtPlotMarker *mY = new QwtPlotMarker();
  mY->setLabelAlignment(Qt::AlignRight|Qt::AlignTop);
  mY->setLineStyle(QwtPlotMarker::HLine);
  mY->setYValue(0.0);
  QPen p = mY->linePen(); 
  p.setColor(Qt::gray);
  mY->setLinePen(p);

  mY->attach(graph);

  QwtPlotGrid *grid = new QwtPlotGrid;
  p.setStyle(Qt::DotLine);
  grid->setPen(p);
  grid->attach(graph);
  

  curve = new QwtPlotCurve("Input Channel");

  curve->setPen(QPen(Qt::white));

  l->addWidget(graph, 0, 0);


  // Set axis titles
  graph->setAxisTitle(QwtPlot::xBottom, "Time/seconds");
  graph->setAxisTitle(QwtPlot::yLeft, "Volts");
  
  curve->attach(graph);
  graph->enableAxis(QwtPlot::xBottom);
  cursorPicker = NULL;

  settings.setPath("GNU", "ComediClient", QSettings::User);
  settings.beginGroup("Comedi Client");
  loadSettings();

  // this replots the graph, etc.. call this after loadSettings..
  applyGraphResize();

  createMenuBar();

  // do connections
  QObject::connect(form->connect, SIGNAL(clicked()), this, SLOT(connect()));
  QObject::connect(form->disconnect, SIGNAL(clicked()), this, SLOT(disconnect()));
  QObject::connect(&dns, SIGNAL(resultsReady()), this, SLOT(resolvedHost()));
  QObject::connect(&dnsTimer, SIGNAL(timeout()), this, SLOT(resolverTimeout()));
  QObject::connect(form->chanAI, SIGNAL(activated(int)), this, SLOT(changeInput(int)));
  QObject::connect(form->gainAI, SIGNAL(activated(int)), this, SLOT(changeInputGain(int)));
  QObject::connect(form->playPause, SIGNAL(toggled(bool)), this, SLOT(pauseUnpause(bool)));
  //QObject::connect(form->voltageAO, SIGNAL(textChanged(const QString &)), this, SLOT(setOutput()));
  QObject::connect(form->voltageAO, SIGNAL(returnPressed()), this, SLOT(setOutput()));
  QObject::connect(form->setVoltBut, SIGNAL(clicked()), this, SLOT(setOutput()));
  QObject::connect(form->chanAO, SIGNAL(activated(int)), this, SLOT(changeOutputGainBox(int)));
  QObject::connect(&keepaliveTimer, SIGNAL(timeout()), this, SLOT(doKeepAlive()));
  setConnected(false);
}

ClientMgr::~ClientMgr()
{
  delete settingsDialog;
  delete cursorPicker;
  delete curve;
  delete graph;  
  disconnect();
  saveSettings();
  settings.endGroup();
}

void ClientMgr::applyGraphResize()
{
  int numPts = inputPeriodHz * graphTimeSecs;
  graphDataX.resize(numPts, 0.0);
  graphDataY.resize(numPts, 0.0);
  // Attach (don't copy) data. 
  curve->setRawData(&graphDataX[0], &graphDataY[0], numPts);

  // X points are static.. set them now
  for (int i = 0; i < numPts; ++i) 
    graphDataX[i] = static_cast<double>(-graphTimeSecs) + static_cast<double>(i)/static_cast<double>(numPts) * graphTimeSecs;

  graph->replot();
}

void ClientMgr::loadSettings()
{
  // populate the servers combobox
  QStringList servers = settings.readListEntry("/servers");  
  for (QStringList::iterator it = servers.begin(); it != servers.end(); ++it) {
    form->serverBox->insertItem(*it);
  }
  settings.beginGroup("/graph"); 
  {
    inputPeriodHz = settings.readNumEntry("/inputPeriodHz", defInputPeriodHz);
    graphTimeSecs = settings.readNumEntry("/graphTimeSecs", defGraphTimeSecs);
    
    if (inputPeriodHz < minInputPeriodHz || inputPeriodHz > maxInputPeriodHz)
      inputPeriodHz = defInputPeriodHz;
    if (graphTimeSecs < minGraphTimeSecs || graphTimeSecs > maxGraphTimeSecs)
      graphTimeSecs = defGraphTimeSecs;
    
  } 
  settings.endGroup();
}

void ClientMgr::saveSettings()
{
  settings.beginGroup("/graph");
  settings.writeEntry("/inputPeriodHz", inputPeriodHz);
  settings.writeEntry("/graphTimeSecs", graphTimeSecs);
  settings.endGroup();

  // save the servers from the combobox
  QStringList servers;  
  for (int i = 0; i < form->serverBox->count(); ++i) 
    servers.push_back(form->serverBox->text(i));
  
  settings.writeEntry("/servers", servers);

}

void ClientMgr::resolvedHost()
{
  dnsTimer.stop();
  if (connecting) {
    QValueList<QHostAddress> addies = dns.addresses();
    host = (*addies.begin()).toString();
    doConnect();
  }
}

void ClientMgr::resolverTimeout()
{
  disconnect();
  dnsTimer.stop();
  form->statusLine->setText(QString("Failed to resolve address for %1").arg(hostName));    
  dns.setLabel(QString::null);
}

void ClientMgr::doConnect()
{
  form->statusLine->setText(QString("Connecting to %1 port %2...").arg(host).arg(port));    
  delete comm;
  comm = new QSocketDevice;
  comm->setBlocking(true);
  if (!comm->connect(host, port)) {
    connectError(comm->error());
    return;
  } 
  connecting = false;
  if (!getServerVersion()) return;
  if (!getDevices()) return; // asks server for all comedi devs, updates UI
  setConnected(true);
  int pos = -1;
  QString userHostPortText = form->serverBox->currentText();
  for (int i = 0; pos < 0 && i < form->serverBox->count(); ++i)
    if (form->serverBox->text(i) == userHostPortText) 
      pos = i; 
  
  if (pos > -1) form->serverBox->removeItem(pos); // remove it if it is there
  // reinsert it at the top to indicate it has just been used
  form->serverBox->insertItem(userHostPortText, 0); 
  form->serverBox->setCurrentItem(0);
  form->serverGB->setTitle(QString("Server: %1").arg(userHostPortText));

  restoreAOAIComboBoxes(); // restore the state of AI/AO combo boxes..
}

void ClientMgr::connectError(int err)
{
    disconnect();
    QString errTxt = SocketMgr::socketErrorString(err);
    form->statusLine->setText(QString("Connection error: <font color=blue>%1</font>").arg(errTxt));
}

void ClientMgr::connect()
{
  disconnect(); // paranoia
  disconnecting = false;
  connecting = true;

  QString serverString = form->serverBox->currentText();
  hostName = serverString;
  port = Protocol::DefaultPort;
  QStringList hostPort = QStringList::split(":", serverString);
  if (hostPort.size() == 2) {
    hostName = hostPort[0];
    bool ok;
    port = hostPort[1].toUShort(&ok);
    if (!ok) port = Protocol::DefaultPort;
  }
  form->statusLine->setText(QString("Resolving %1...").arg(hostName));  

  form->serverGB->setEnabled(false);
  form->serverSelection->setEnabled(true);
  form->connect->setEnabled(false);
  form->disconnect->setEnabled(true);
  form->serverBox->setEnabled(false);
  hostName = hostName.simplifyWhiteSpace();
  if ( hostName.lower() == "localhost" 
       || hostName.lower() == "localhost.localdomain" ) 
    hostName = "127.0.0.1"; // hack because QDns is broken...
  if ( hostName.find(QRegExp("^\\s*\\d{1,3}[.]\\d{1,3}[.]\\d{1,3}[.]\\d{1,3}\\s*$")) >
-1) {
    // if it's already in dot notation, just do the connection right away, skipping the  resolver step
      host = hostName;
      doConnect();
  } else {
    dnsTimer.start(10000, true);
    dns.setRecordType(QDns::A);
    dns.setLabel(hostName);
  }  
}

template<typename T> void deleteIt(T it) {  delete it; }

void ClientMgr::disconnect()
{
  saveAOAIComboBoxes();

  dnsTimer.stop();    
  keepaliveTimer.stop();
  delete comm; comm = 0;
  serverVersion = 1;
  exclusiveDIOs.clear();
  disconnecting = true;
  connecting = false;
  delete [] devices;
  devices = 0;
  std::for_each(datas.begin(), datas.end(), deleteIt<SocketMgr*>);
  datas.clear();
  data = NULL;
  sockStateMap.clear();
  devices = NULL;
  devices = NULL;
  numDevices = 0;
  cidcb_in.clear();
  cbcid_in.clear();
  cidcb_out.clear();
  cbcid_out.clear();
  smcid.clear();
  cidsm.clear();
  form->chanAO->clear();
  form->chanAI->clear();
  form->gainAO->clear();
  form->gainAI->clear();
  form->serverGB->setTitle(QString("Server: xxxx"));
  setConnected(false);
  // forget the host, hostName, etc
  host = "";
  hostName = "";
}

void ClientMgr::removeSocketMgr(SocketMgr *s)
{
  if (sockStateMap.find(s) != sockStateMap.end()) {
    for (Datas::iterator dit = datas.begin(); dit != datas.end(); ++dit)
      if (*dit == s) { datas.erase(dit); break; }
    sockStateMap.erase(s);
    delete s;
  }
  SMCID::iterator it;
  if ( (it = smcid.find(s)) != smcid.end() ) {
    CIDEx c = smcid[s];
    cidsm.erase(c);
    smcid.erase(s);
  }
}

void ClientMgr::setConnected(bool con)
{
  form->serverGB->setEnabled(con);
  form->serverBox->setEnabled(!con);
  form->connect->setEnabled(!con);
  form->disconnect->setEnabled(con);
  if (con) {
    form->statusLine->setText(QString("Connected to %1").arg(form->serverBox->currentText()));
  } else {
    form->statusLine->setText("<font color=red>Not connected</font>");
  }
}

bool ClientMgr::getDevices()
{
  if ( !sendCommLine(Protocol::GetComediDevices) ) {
    disconnect();
    form->statusLine->setText("<font color=red>Send error getting devices.</font>");
    return false;
  }
  QByteArray resp;
  resp.resize(sizeof(int));
  //if (comm->bytesAvailable() < resp.size()) comm->waitForMore(10000); // eek, block here
  if ( comm->readBlock(resp.data(), resp.size()) != (int)resp.size() ) {
    disconnect();
    form->statusLine->setText("<font color=red>Receive error getting device count.</font>");
    return false;
  }
  delete [] devices;
  devices = 0;
  le_to_dw(resp.data()); // convert transmitted binary 32-bit int from protocol format to machine format
  int num = *reinterpret_cast<int *>(resp.data());
  
  if (num >= 16 || num < 1) // allow no more than 16 comedi devices 
    {
          disconnect();
          form->statusLine->setText("<font color=red><b>Disconnected: invalid number of comedi devices coming in!</b></font>");
          return false;
    }
  devices = new ComediDevice[ num ];
  long sz = num*sizeof(ComediDevice), xferred = 0;
  resp.resize(sz);
  while (xferred < sz) {
    int nread, n2read = sz-xferred;
    if (!comm->bytesAvailable()) comm->waitForMore(2000);
    if ( (nread = comm->readBlock(resp.data(), n2read)) <= 0 ) {
      disconnect();
      form->statusLine->setText("<font color=red><b>Disconnected: invalid number of bytes coming in!</b></font>");
      return false;
    }
    ::memcpy(((char *)devices) + xferred, resp.data(), nread);
    xferred += nread;
  }
  numDevices = num;
  comediDeviceFromLE(devices, numDevices); // convert it from protocol order (little endian) to machine order (if we already are little endian.. this is a noop)
  if ( !gotDevices() ) return false;
  if (!startAllDIOACQs()) {
    form->statusLine->setText("<font color=red><b>Disconnected: failure in initiating data acquisition connections to server!</b></font>");
    return false;
  }
  return true;
}

bool ClientMgr::getServerVersion() // asks server for its version number -- some servers don't understand this command -- side effect is to set serverVersion.  serverVersion is set to 1 if the server doesn't uinderstand this commant
{
  QString line = Protocol::GetVersion;
  
  if (!sendCommLine(line)) {
    disconnect();
    form->statusLine->setText("<font color=red><b>Disconnected: failure in querying server version!</b></font>");
    return false;
  }
  line = receiveCommLine();
  if (line.isNull()) {
    disconnect();
    form->statusLine->setText("<font color=red><b>Disconnected: failure in querying server version!</b></font>");
    return false;    
  }

  // so far so good, we successfully sent a command to the server and
  // read a line back.

  // now.. see if the server version was returned or if the server is too
  // old and doesn't understand this command..
  bool ok = false;
  serverVersion = line.toUInt(&ok);
  if (!ok) {
    serverVersion = 1;
    delete comm;
    comm = new QSocketDevice;    
    comm->connect(host, port);
    // reconnect since at this point server may have closed connection on us
    // if they didn't support the protocol command!!
  }
  return true;
}


bool ClientMgr::gotDevices()
{
  cidcb_in.clear();
  cbcid_in.clear();
  cidcb_out.clear();
  cbcid_out.clear();
  form->chanAO->clear();
  form->chanAI->clear();
  minorMap.clear();

  std::list<CID> dioSubdevs;

  for (unsigned i = 0; i < numDevices; ++i) {
    ComediDevice *dev = devices + i;
    minorMap[dev->minor] = dev;
    for (unsigned j = 0; j < dev->nSubdevs; ++j) {
      ComediSubDevice *sdev = dev->subdev + j;

      /* Remember the DIO subdevs for later on when we have to query
         their DIOConfig state below.. */

      if (sdev->type == ComediSubDevice::DIO) {
        CID c;
        c.dev = dev->minor;
        c.subdev = sdev->id;
        c.chan = 0;
        dioSubdevs.push_back(c);
      }
          
      for (unsigned k = 0; k < sdev->nChans; ++k) {
        QString str("comedi%1 - %2%3-%4");
        QString sdevstr;
        switch(sdev->type) {
        case ComediSubDevice::AI: sdevstr = "AI"; break;
        case ComediSubDevice::AO: sdevstr = "AO"; break;
        case ComediSubDevice::DI: sdevstr = "DI"; break;
        case ComediSubDevice::DO: sdevstr = "DO"; break;
        case ComediSubDevice::DIO: sdevstr = "DIO"; break;
        default: sdevstr = "??"; break;
        }
        str = str.arg(dev->minor).arg(sdevstr).arg(sdev->id).arg(k);
        CIDEx c;
        c.dev = i; 
        c.subdev = sdev->id;
        c.chan = k;

        if (sdev->type == ComediSubDevice::AI 
            || sdev->type == ComediSubDevice::DI
            || sdev->type == ComediSubDevice::DIO
            || sdev->type == ComediSubDevice::AO ) {
          form->chanAI->insertItem(str);
          int index = form->chanAI->count()-1;
          cidcb_in[c] = index;
          cbcid_in[index] = c;
        }
        if (sdev->type == ComediSubDevice::AO 
            /*|| sdev->type == ComediSubDevice::DO
              || sdev->type == ComediSubDevice::DIO */) {
          form->chanAO->insertItem(str);        
          int index = form->chanAO->count()-1;
          cidcb_out[c] = index;
          cbcid_out[index] = c;
        }

        // exclusive DIO set setup.. for special olifactometer control
        // channels 0-15 are mutually exclusive..
        if ((sdev->type == ComediSubDevice::DIO
             || sdev->type == ComediSubDevice::DO)
            && c.chan <= 15) {
          exclusiveDIOs.insert(c);
        }
      }
    }
  }

  if ( !changeInput(form->chanAI->currentItem()) ) return false;
  // redo the diomgr
  delete dioMgr;
  delete scrollView;
  if (form->dioGB->layout()) delete form->dioGB->layout();
  QGridLayout *l = new QGridLayout(form->dioGB, 1, 1);
  scrollView = new QScrollView(form->dioGB);
  l->addWidget(scrollView, 0, 0);
 
  dioMgr = new DIOMgr(devices, numDevices, scrollView->viewport(), "DIO Manager");
  // now query the DIO configs to setup the checkboxes to reflect server state
  for (std::list<CID>::iterator it = dioSubdevs.begin(); it != dioSubdevs.end(); ++it) {
    CID c = *it;
    ConfigVec v = dioGetConfig(c.dev, c.subdev);
    for (c.chan = 0; c.chan < v.size(); ++c.chan) {      
      dioMgr->setDIOCB(c, v[c.chan]);
    }
  }
  scrollView->addChild(dioMgr);
  QObject::connect(dioMgr, SIGNAL(dioConfigModifyRequested(CID, ComediSubDevice::DIOConfig)), this, SLOT(doDIOConfigModify(CID, ComediSubDevice::DIOConfig)));
  QObject::connect(dioMgr, SIGNAL(dioWriteRequested(CID, bool)), this, SLOT(doDIOWrite(CID, bool)));
  scrollView->clipper()->setBackgroundMode(Qt::PaletteBackground); // make the background of the canvas 'invisible'
  scrollView->show();  
  return true;
}


void ClientMgr::recvFinished(const QByteArray & data_array)
{  
  if (disconnecting) return; // ignore spurious/leftover signals?

  QObject *o = const_cast<QObject *>(sender());
  SocketMgr *mgr = o ? dynamic_cast<SocketMgr *>(o) : 0;
  
  if (!mgr) {
    disconnect();
    form->statusLine->setText(QString("<font color=red><b>Disconnected: Unexpected SocketMgr event in ClientMgr::recvFinished()!</font>"));
    return;
  }
  SMCID::iterator it = smcid.find(mgr);
  if (it == smcid.end()) {
    std::cout << "DEBUG: deleted/removed data came in recvFinished!\n";
    return;
  }
  CIDEx c = it->second;

  SockState & stateRef = sockStateMap[mgr].first;

  switch (stateRef) {
  case ReceivingData:
    {
      /*Sample_t samp;
      QByteArray arr;      
      arr.resize(data->bytesAvailable());
      data->readBlock(arr.data(), arr.size()); */
      int num = data_array.size() / sizeof(Sample_t);

      // DEBUG
      //std::cerr << "got " << num << " samples for " << c << "\n";

      Sample_t *data = reinterpret_cast<Sample_t *>(data_array.data());
      
      if (!c.isAllDIO) {

        if (currentRange && !graphPaused && num) {
          int numPlotPoints = num;
          if (numPlotPoints > (int)graphDataX.size()) 
            numPlotPoints = graphDataX.size();
          double *front = &graphDataY[0], *pos = &graphDataY.back() - numPlotPoints + 1;
          ::memmove(front, front + numPlotPoints, (graphDataY.size()-numPlotPoints)*sizeof(double));
          for (int i = 0; i < numPlotPoints; ++i) {
            sampleFromLE(&data[i]); // convert from protocol order to machine order
            if (data[i] > currentMaxData) data[i] = (unsigned)currentMaxData; // note how this forces DIO channels to be 1/0 rather than the bitmask they come in as
            pos[i] = currentRange->min + ((currentRange->max - currentRange->min) * static_cast<double>(data[i])/currentMaxData);
          }
          // update the LCD
          form->voltageAI->display(graphDataY.back());
//         graph->setData(0, graphDataX.size(), &graphDataX[0], &graphDataY[0]);
//         graph->repaint(false);

          // raw data is already attached..
          //curve->setData(&graphDataX[0], &graphDataY[0], graphDataX.size());
          graph->replot();
        }

      } else {
        sampleFromLE(&data[num-1]); // convert from protocol order to machine order
        gotDIOData(c, data[num-1]);
      }

    }
    break;

  default:
    disconnect();
    form->statusLine->setText(QString("<font color=red><b>Disconnected: Unexpected socket state in recvFinished!</font>"));
    return;
    break;
  }
}

void ClientMgr::sendFinished(int numBytes)
{  
  if (disconnecting) return; // ignore spurious/leftover signals?

  (void)numBytes;
  QObject *o = const_cast<QObject *>(sender());
  SocketMgr *mgr = o ? dynamic_cast<SocketMgr *>(o) : 0;

  if (!mgr) {
    disconnect();
    form->statusLine->setText(QString("<font color=red><b>Disconnected: Unexpected SocketMgr event in ClientMgr::sendFinished()!</font>"));
    return;
  }

  SockState & stateRef = sockStateMap[mgr].first;
  //SockState & nextStateRef = sockStateMap[mgr].second;

  switch(stateRef) {
  default:
    disconnect();
    form->statusLine->setText(QString("<font color=red>Disconnected: Unexpected socket state in sendFinished()!</font>"));
    return;
    break;
  }
}

void ClientMgr::populateGains()
{
  form->gainAO->clear();
  form->gainAI->clear();
  int index;
  // do AI
  if (form->chanAI->count() && (index = form->chanAI->currentItem()) > -1) {
    CIDEx c = cbcid_in[index];
    ComediSubDevice & sdev = devices[c.dev].subdev[c.subdev];
    for (unsigned r = 0; r < sdev.nRanges; ++r) {
      ComediRange *range = sdev.range + r;
      form->gainAI->insertItem(QString("%1V-%2V").arg(range->min).arg(range->max));
    }
  }
  // do AO
  if (form->chanAO->count() && (index = form->chanAO->currentItem()) > -1) {
    CIDEx c = cbcid_out[index];
    ComediSubDevice & sdev = devices[c.dev].subdev[c.subdev];
    for (unsigned r = 0; r < sdev.nRanges; ++r) {
      ComediRange *range = sdev.range + r;
      form->gainAO->insertItem(QString("%1V-%2V").arg(range->min).arg(range->max));
    }
  }
}

bool ClientMgr::changeInput(int item)
{
  int oldGain = form->gainAI->currentItem(), oldGain2 = form->gainAO->currentItem();
  populateGains();
  if (oldGain > -1 && oldGain < form->gainAI->count()) 
    form->gainAI->setCurrentItem(oldGain);
  if (oldGain2 > -1 && oldGain2 < form->gainAO->count()) 
    form->gainAO->setCurrentItem(oldGain2);
  int gain = form->gainAI->currentItem();

  if (data) 
    removeSocketMgr(data);
  data = NULL;
  
  CIDEx c = cbcid_in[item];
  if (!createDataSocket(c, gain)) // NB: this method sets appropriate error text
    return false;
  data = cidsm[c];
  
  if (gain >= 0) {
    ComediDevice *dev = findDeviceByMinor(c.dev);
    ComediSubDevice *subdev = dev->subdev + c.subdev;
    if (gain < (int)subdev->nRanges) {
      currentRange = subdev->range + gain;
      currentMaxData = subdev->maxdata;
      graph->setAxisScale(QwtPlot::yLeft, currentRange->min, currentRange->max);
    }
  }
  return true;
}

bool ClientMgr::changeInputGain(int) // param is index into form->gainAI
{
  // we have to basically restart the acq.
  return changeInput(form->chanAI->currentItem());
}

void ClientMgr::doSocketMgrSignalConnections(SocketMgr *s)
{
  QObject::connect(s, SIGNAL(dataSent(int)), this, SLOT(sendFinished(int)));
  QObject::connect(s, SIGNAL(socketError(int)), this, SLOT(gotSocketError(int)));
  QObject::connect(s, SIGNAL(dataReceived(const QByteArray &)), this, SLOT(recvFinished(const QByteArray &)));
  QObject::connect(s, SIGNAL(socketError(int)), this, SLOT(gotSocketError(int)));  
}

void ClientMgr::gotSocketError(int err)
{
  if (disconnecting) return; // ignore spurious/leftover signals?
  disconnect();
  form->statusLine->setText(QString("<font color=red>Disconnected: %1</font>").arg(SocketMgr::socketErrorString(err)));
}

bool ClientMgr::startAllDIOACQs()
{
  for (ComediDevice * dev = devices; dev < devices + numDevices; ++dev)
    for (ComediSubDevice *sdev = dev->subdev; sdev < dev->subdev + dev->nSubdevs; ++sdev)
      if (sdev->type == ComediSubDevice::DIO
          || sdev->type == ComediSubDevice::DI 
          || sdev->type == ComediSubDevice::DO) {
        CIDEx c;
        c.dev = dev->minor;
        c.subdev = sdev->id;
        c.chan = 0; 
        c.isAllDIO = true; // special ALL DIO acquisition
        if (!createDataSocket(c)) return false;
      }
  return true;
}

bool ClientMgr::createDataSocket(CIDEx c, unsigned gain, unsigned hz)
{
  QString bitmask = "";  
  if (c.isAllDIO) { // if it's a DIO 'all channels' bitfield acq.
    hz = dioInputPeriodHz;
    ComediDevice *dev = findDeviceByMinor(c.dev);
    if (!dev) {
      std::cerr << "FATAL: findDeviceByMinor() returned NULL!\n";
      ::abort();
    }
    ComediSubDevice *sdev = dev->subdev + c.subdev;
    unsigned mask = (0x1 << sdev->nChans) - 1;
    if (sdev->nChans >= 32) mask = 0xffffffff;
    bitmask = QString("%1").arg(mask);
    if (!hz) hz = dioInputPeriodHz;
  }
  if (!hz) hz = inputPeriodHz;
  QString line = QString("%1 %2 %3 %4 %5 %6\n").arg(Protocol::StartACQ)
    .arg(c).arg(gain).arg(0).arg(hz).arg(bitmask).latin1();

  if (!sendCommLine(line)) {
    disconnect();
    form->statusLine->setText(QString("<font color=red>Disconnected: Error writing to control socket!</font>"));
    return false;
  }
  line = receiveCommLine();
  if ( line.isNull()|| line.startsWith(Protocol::ErrorText, false) ) {
    disconnect();
    form->statusLine->setText(QString("<font color=red>Disconnected: Unexpected response: '%1'!</font>").arg(line));
    return false;   
  }
  
  QSocketDevice *sock = new QSocketDevice;
  sock->setBlocking(true);
  if (!sock->connect(host, port)) {
    disconnect();
    form->statusLine->setText(QString("<font color=red>Error connecting data socket: %1!</font>").arg(SocketMgr::socketErrorString(sock->error())));
    delete sock; 
    return false;
  }
  line = QString(Protocol::YouAreADataThread) + " " + line;
  if ( !sendLine(sock, line) ) {
    disconnect();
    form->statusLine->setText(QString("<font color=red>Error negotiating data connection: %1!</font>").arg(SocketMgr::socketErrorString(sock->error())));
    delete sock;
    return false;
  }
  line = receiveLine(sock);
  if ( line.isNull() || !line.startsWith(Protocol::OKText) ) {
    delete sock;
    disconnect();
    form->statusLine->setText(QString("<font color=red>Disconnected: Unexpected data socket response in createDataSocket(): '%1'!</font>").arg(line));
    return false;
  }
  sock->setBlocking(false);
  SocketMgr *s = new SocketMgr(SocketMgr::Read, sock, this, QString("Data %1 Socket Manager").arg(c).latin1());
  datas.push_back(s);
  doSocketMgrSignalConnections(s);
  cidsm[c] = s;
  smcid[s] = c;
  s->requestData(sizeof(Sample_t));
  sockStateMap[s].first = ReceivingData;
  sockStateMap[s].second = ReceivingData;
  return true;
}

ComediDevice *ClientMgr::findDeviceByMinor(unsigned minor) const
{
  return const_cast<ClientMgr *>(this)->minorMap[minor];
}

void ClientMgr::pauseUnpause(bool tog) 
{
  graphPaused = tog;
}

void ClientMgr::gotDIOData(CIDEx c, unsigned int chanBits)
{
  if (!dioMgr) return;
  ComediDevice *dev = findDeviceByMinor(c.dev);
  ComediSubDevice & sdev = dev->subdev[c.subdev];
  for (unsigned i = 0; i < sdev.nChans; ++i) {
    c.chan = i;
    dioMgr->setDIOState(c, chanBits & (0x1<<i));
  }
}

void ClientMgr::setOutput()
{
  QString str = form->voltageAO->text();
  CIDEx c = cbcid_out[form->chanAO->currentItem()];
  int gain = form->gainAO->currentItem();
  bool ok;
  double v = str.toDouble(&ok);
  if (!ok) return; // parse error on double value...
  ComediDevice *dev = findDeviceByMinor(c.dev);
  ComediSubDevice *sdev = dev->subdev + c.subdev;
  ComediRange *range = sdev->range + gain;
  
  if (sdev->type == ComediSubDevice::DIO)
    dioMgr->setDIOCB(c, ComediSubDevice::Output); // reflect that it's an output now in the DIO display..

  // compute sample from range setting
  Sample_t sample = static_cast<Sample_t>((v - range->min)/(range->max - range->min) * sdev->maxdata);

  if (v <= range->min) {
    sample = 0;
    v = range->min;
  } else if (v >= range->max) {
    sample = sdev->maxdata;
    v = range->max;
  }

  // digitize DIO.. either 1 or 0
  if (sdev->type == ComediSubDevice::DIO || sdev->type == ComediSubDevice::DO) {
    if (v > 0.0) {
      sample = sdev->maxdata;
      v = 5.0;
    } else {
      v = 0.0;
      sample = 0;
    }

  }
  
  QString line = QString("%1 %2 %3 %4 %5\n").arg(Protocol::WriteData).arg(c).arg(sample).arg(gain).arg(0);
  
  // send command and receive response.. if that fails.. disconnect
  if (!sendCommLine(line)) {
    disconnect();
    form->statusLine->setText("<font color=red>Disconnected: could not set output line</font>");
    return;
  }
  line = receiveCommLine();
  if ( line.isNull() || !line.startsWith(Protocol::OKText, false) ) { 
    disconnect();
    form->statusLine->setText("<font color=red>Disconnected: could not set output line</font>");
    return;
  } 

  form->statusLine->setText(QString("Set output on %1 to %2").arg(form->chanAO->currentText()).arg(v));    
  
}

void ClientMgr::changeOutputGainBox(int item)
{
  (void) item;
  int oldGain = form->gainAI->currentItem(), oldGain2 = form->gainAO->currentItem();
  populateGains();
  if (oldGain > -1 && oldGain < form->gainAI->count()) 
    form->gainAI->setCurrentItem(oldGain);
  if (oldGain2 > -1 && oldGain2 < form->gainAO->count()) 
    form->gainAO->setCurrentItem(oldGain2);
}



class DIOChk : public QCheckBox, public CIDRememberer
{
public:
  DIOChk(QWidget *parent = 0, const char *name = 0) : QCheckBox(parent, name){}
};


DIOLed::DIOLed(QWidget *p, const char *n)
  : QWidget(p, n), ishigh(true), high("#ff0000"), low("#661111")
{
  setMaximumWidth(10);
  setMaximumHeight(10);
}

DIOLed::~DIOLed() {}

void DIOLed::setState(bool h)
{
  ishigh = h;
  repaint();
}

void DIOLed::paintEvent(QPaintEvent *e)
{
  (void)e;
  QPainter p(this);
  QColor color = ishigh ? high : low;

  // gray it out if it is disabled..
  if (!isEnabled()) color = QColor((Qt::gray.red()+color.red())/2,
                                   (Qt::gray.green()+color.green())/2,
                                   (Qt::gray.blue()+color.blue())/2);
  p.setBrush( Qt::SolidPattern );
  p.setBrush( color );
  p.setPen( color );
  p.drawRect(0, 0, width(), height());
}

void DIOLed::mousePressEvent(QMouseEvent *e)
{
  e->accept();
  emit clicked();
}


DIOMgr::DIOMgr(ComediDevice *devices, unsigned numDevs,
               QWidget *parent, const char *name)
  : QWidget(parent, name)
{
  std::list<ComediSubDevice *> dioSubDevs;
  
  // compute dio subdevs, set up the minor map, etc 
  int dioSubdevCt = 0;

  for (unsigned i = 0; i < numDevs; ++i) {
    for (unsigned s = 0; s < devices[i].nSubdevs; ++s) {
      if (devices[i].subdev[s].type == ComediSubDevice::DIO
          || devices[i].subdev[s].type == ComediSubDevice::DI
          || devices[i].subdev[s].type == ComediSubDevice::DO) {
        dioSubDevs.push_back(&devices[i].subdev[s]);
        ++dioSubdevCt;
      }
    }
  }
  
  QGridLayout *l = new QGridLayout(this, dioSubdevCt*3, 33, 0, 0, "DIO Button Grid");
  int row = 0;
  for (std::list<ComediSubDevice *>::iterator s = dioSubDevs.begin();
       s != dioSubDevs.end();
       ++s, ++row)
  {
    if (row) {
      QSpacerItem *spacer = new QSpacerItem(1, 7); // spacer between pairs of DIO/config rows..
      l->addItem(spacer, row++, 0);
    }
    ComediSubDevice::Type type = (ComediSubDevice::Type) (*s)->type;
    QString typeStr;
    bool isLedClickable = true;
    switch (type) {
    case ComediSubDevice::DIO: typeStr = "DIO"; break;
    case ComediSubDevice::DI: typeStr = "DI"; isLedClickable = false; break;
    case ComediSubDevice::DO: typeStr = "DO"; break;
    default: typeStr = "???"; break;
    }
    QLabel * label = new QLabel(QString("comedi%1-%2-%3:").arg((*s)->minor).arg(typeStr).arg((*s)->id), this);
    l->addWidget(label, row, 0);
    label->setAlignment(Qt::AlignRight);
    if (type == ComediSubDevice::DIO) {
      label = new QLabel("<font size=-1>(chk for output)</font>:", this);
      label->setAlignment(Qt::AlignRight);
      l->addWidget(label, row+1, 0);
    }
    for (unsigned chan = 0; chan < (*s)->nChans; ++chan) {
      CID c;
      c.chan = chan; c.dev = (*s)->minor; c.subdev = (*s)->id;
      DIOLed *but = new DIOLed(this);
      but->setCID(c); // make the DIOButton remember its own CID!
      buttonMap[c] = but;
      l->addWidget(but, row, chan+1);
      setDIOState(c, false);
      if (isLedClickable)
        connect(but, SIGNAL(clicked()), this, SLOT(doDioToggle()));
      
      if (type == ComediSubDevice::DIO) {
        DIOChk *chk = new DIOChk(this);
        chk->setCID(c); // make the checkbox remember its own CID.
        chk->setChecked(false);
        QToolTip::add(chk, QString("Check to set channel %1 for OUTPUT").arg(c.chan));
        checkBoxMap[c] = chk;
        l->addWidget(chk, row+1, chan+1);
        connect(chk, SIGNAL(toggled(bool)), this, SLOT(doDioChecked(bool)));
      }
    }
    ++row;
  }
}

DIOMgr::~DIOMgr(){}

void DIOMgr::setDIOState(CID c, bool high)
{
      DIOLed *but = buttonMap[c];
      but->setState(high);
      QString word1 = "low", word2 = "high";
      if (high) word1 = "high", word2 = "low";
      QToolTip::remove(but);
      QToolTip::add(but, QString("Line %1 is now %2.  Click to toggle is back %3.").arg(c.chan).arg(word1).arg(word2));
}

void DIOMgr::doDioToggle()
{
  const DIOLed *l;
  CID c =  (l = dynamic_cast<const DIOLed *>(sender()))->getCID();
  bool newstate = !l->state();
  //setDIOState(c, newstate); // make the color change take place
  emit dioWriteRequested(c, newstate);
}

void DIOMgr::doDioChecked(bool output)
{
  CID c =  dynamic_cast<const CIDRememberer *>(sender())->getCID();
  QCheckBox *chk = const_cast<QCheckBox *>(dynamic_cast<const QCheckBox *>(sender()));
  if (chk) {
    QToolTip::remove(chk);
    if (output)
      QToolTip::add(chk, QString("Un-check to set channel %1 for INPUT").arg(c.chan));
    else
      QToolTip::add(chk, QString("Check to set channel %1 for OUTPUT").arg(c.chan));
  }  

  emit dioConfigModifyRequested(c, output ? ComediSubDevice::Output : ComediSubDevice::Input);
}

void ClientMgr::doDIOConfigModify(CID c, ComediSubDevice::DIOConfig conf)
{
  doOkCmd( QString("%1 %2 %3\n").arg(Protocol::DIOConfig).arg(c).arg((unsigned int)conf), "DIO conf. modify" );
}

void ClientMgr::doDIOWrite(CID c, bool bit)
{
  unsigned int data = bit, chanMask = 0;

  // special exclusive-mode hack.  Check our set of exclusive channels
  // and, based on membership in that set, do a bitfield DIO
  // setting all exclusive channels to 0 and setting the one they clicked
  // to 1.  Note we ignore the actual value of bit in this case
  if (form->exclusiveBut->isChecked() && 
      exclusiveDIOs.find(c) != exclusiveDIOs.end()) {
    data = 0x1 << c.chan; // set only the c.chan'th bit high...
    // now setup the chan write mask based on membership in the exclusiveDIOs set..
    for (int i = 0; i < 32; ++i) {
      c.chan = i;
      // indicate to comedi that all the exlusive channels should be 
      // written-to.  Basically all our exclusive channels are in the 
      // write mask (chanmask)
      if (exclusiveDIOs.find(c) != exclusiveDIOs.end())
        chanMask |= 0x1 << i;
    }
  }

  doOkCmd( QString("%1 %2 %3 0 0 %4\n").arg(Protocol::WriteData).arg(c).arg(data).arg(chanMask), "writing DIO");

}

void ClientMgr::dioConfigSubdev(unsigned dev, unsigned subdev, 
                                ComediSubDevice::DIOConfig conf)
{
  unsigned mask = 0xffffffff;
  QString line;
  CID c;
  c.dev = dev;
  c.subdev = subdev;
  c.chan = 0; // ignored

  doOkCmd( QString("%1 %2 %3 %4\n").arg(Protocol::DIOConfig).arg(c).arg((unsigned int)conf).arg(mask), "configure DIO" );
}

ClientMgr::ConfigVec
ClientMgr::dioGetConfig(unsigned dev, unsigned subdev)  
{
  ClientMgr::ConfigVec vec;
  ComediDevice *d = findDeviceByMinor(dev);
  vec.resize(d->subdev[subdev].nChans, ComediSubDevice::Input);
  if (serverVersion > 1) {
    u32 mask = 0xffffffff;
    CID c;
    c.dev = dev;
    c.subdev = subdev;
    c.chan = 0; // ignored
    QString line = QString("%1 %2 %3\n").arg(Protocol::GetDIOConfig).arg(c).arg(mask);
    
    if (sendCommLine(line)) {
      // read back response line, parse results, put them in vector
      line = receiveCommLine();
      if (!line.isNull()) {
        std::istringstream iss(line.latin1());
        int i = 0, tmp;
        while (iss >> tmp && i < (int)vec.size()) {
          vec[i++] = static_cast<ComediSubDevice::DIOConfig>(tmp);
        }
      }
    } 
  }
  return vec;
}

void DIOMgr::setDIOCB(CID chan, ComediSubDevice::DIOConfig conf)
{
  QCheckBox *chk = checkBoxMap[chan];
  if (chk) {
    chk->blockSignals(true);
    if (conf == ComediSubDevice::Output) chk->setChecked(true);
    else if (conf == ComediSubDevice::Input) chk->setChecked(false);
    chk->blockSignals(false);
  }
}

/* static */
bool ClientMgr::sendLine(QSocketDevice *s, const QString & l)
{
  QString ll(l);
  if (!ll.endsWith("\n")) ll += "\n";
  QCString line = ll.utf8();
  return s && s->writeBlock(line.data(), line.length()) == (int)line.length(); 
}

QString ClientMgr::receiveLine(QSocketDevice *s)
{
  if (!s) return QString::null;
  QCString line;
  line.resize(1024);
  Q_LONG nread = s->readLine(line.data(), line.size());
  if (nread < 0) return QString::null;
  if (nread >= 0 && nread < (Q_LONG)line.size()) line[nread] = 0;
  line[line.size()-1] = 0;
  return line;
}

bool ClientMgr::sendCommLine(const QString & l)
{
  keepaliveTimer.start(60*1000); // 1 min keepalive timer
  return sendLine(comm, l);
}

QString ClientMgr::receiveCommLine()
{
  keepaliveTimer.start(60*1000); // 1 min keepalive timer
  return receiveLine(comm);
}

bool ClientMgr::doOkCmd(const QString &cmd, const QString & cmd_descr)
{
  if (!sendCommLine(cmd)) {
    disconnect();
    form->statusLine->setText(QString("<font color=red>Error with command '%1': %2</font>").arg(cmd_descr).arg(SocketMgr::socketErrorString(comm->error())));
    return false;
  }
  QString reply = receiveCommLine();
  if ( !reply.startsWith(Protocol::OKText, false) ) {
    disconnect();
    form->statusLine->setText(QString("<font color=red>Error with command '%1':  %2</font>").arg(cmd_descr).arg(reply));
    return false;
  }
  return true;
}

void ClientMgr::saveAOAIComboBoxes()
{
  // cache last host info incase of disconnect/reconnect cycle to same host
  if (form->chanAO->count() && form->chanAI->count()) {
    SavedComboBoxIdxes & s = savedCBStates[host];
    s.out = form->chanAO->currentItem();
    s.in = form->chanAI->currentItem();
    s.outGain = form->gainAO->currentItem();
    s.inGain = form->gainAI->currentItem();
  }
}

void ClientMgr::restoreAOAIComboBoxes()
{
  if (savedCBStates.find(host) != savedCBStates.end()) { // revive cached gain/output combo boxes for this host..
    const SavedComboBoxIdxes & s = savedCBStates[host];
    if (form->chanAI->count() > s.in) {
      form->chanAI->setCurrentItem(s.in);
      changeInput(s.in);
      form->gainAI->setCurrentItem(s.inGain);
      changeInputGain(s.inGain);
    }
    if (form->chanAO->count() > s.out) {
      form->chanAO->setCurrentItem(s.out);
      form->gainAO->setCurrentItem(s.outGain);
      changeOutputGainBox(s.outGain);
    }
  }
}

void ClientMgr::doKeepAlive()
{
  if (serverVersion > 1 && comm && comm->isOpen()) 
    doOkCmd(Protocol::Noop, "keepalive noop");  
}


void ClientMgr::createMenuBar()
{
  QMenuBar *menu = new QMenuBar(form, "ClientUI Menu Bar");

  QPopupMenu *file = new QPopupMenu(menu);
  file->insertItem("&Settings", this, SLOT(doSettingsDialog()));
  file->insertSeparator();
  file->insertItem("&Quit", form, SLOT(close()));
  menu->insertItem("&File", file);
  QPopupMenu *help = new QPopupMenu(menu);
  help->insertItem("&About", this, SLOT(doAboutDialog()));
  help->insertItem("About &Qt", qApp, SLOT(aboutQt()));
  menu->insertItem("&Help", help);
  form->setMenuBar(menu);  
}


SettingsDialog::SettingsDialog(int inputperiod, 
                               int minInputPeriodHz, int maxInputPeriodHz,
                               int graphtime,
                               int minGraphTimeSecs, int maxGraphTimeSecs,
                               QWidget *parent, const char *name, bool modal)
  : QDialog(parent, name, modal)
{
  orig_inputperiod = inputperiod;
  orig_graphtime = graphtime;
  min_inp = minInputPeriodHz;
  max_inp = maxInputPeriodHz;
  min_gt = minGraphTimeSecs;
  max_gt = maxGraphTimeSecs;
  QGridLayout *layout = new QGridLayout(this, 3, 2);
  QLabel *hzLabel = new QLabel("Input sampling rate (Hz): ", this);
  layout->addWidget(hzLabel, 0, 0);
  QIntValidator val1(this);
  val1.setRange(minInputPeriodHz, maxInputPeriodHz);
  inputPeriodLe = new QLineEdit(QString::number(inputperiod), this);
  inputPeriodLe->setValidator(&val1);
  layout->addWidget(inputPeriodLe, 0, 1);

  QLabel *graphTimeLabel = new QLabel("Number of seconds to graph: ", this);
  layout->addWidget(graphTimeLabel, 1, 0);
  QIntValidator val2(this);
  val2.setRange(minGraphTimeSecs, maxGraphTimeSecs);
  graphTimeLe = new QLineEdit(QString::number(graphtime), this);
  graphTimeLe->setValidator(&val2);
  layout->addWidget(graphTimeLe, 1, 1);
  
  QWidget *dummy = new QWidget(this);
  layout->addMultiCellWidget(dummy, 2, 2, 0, 2);
  
  QGridLayout *l2 = new QGridLayout(dummy, 1, 3, 2, 5);

  QPushButton *ok = new QPushButton("Ok", dummy);
  QPushButton *cancel = new QPushButton("Cancel", dummy);
  apply = new QPushButton("Apply", dummy);
  l2->addWidget(ok, 0, 0, Qt::AlignCenter);
  l2->addWidget(cancel, 0, 1, Qt::AlignCenter);
  l2->addWidget(apply, 0, 2, Qt::AlignCenter);
  QObject::connect(ok, SIGNAL(clicked()), this, SLOT(accept()));
  QObject::connect(apply, SIGNAL(clicked()), this, SIGNAL(applyClicked()));
  QObject::connect(apply, SIGNAL(clicked()), this, SLOT(disableApply()));
  QObject::connect(cancel, SIGNAL(clicked()), this, SLOT(reject()));  

  // by default apply button is disabled until something changes..
  apply->setEnabled(false);
  QObject::connect(graphTimeLe, SIGNAL(textChanged(const QString &)), this, SLOT(enableApply()));
  QObject::connect(inputPeriodLe, SIGNAL(textChanged(const QString &)), this, SLOT(enableApply()));
}

int SettingsDialog::inputPeriod() const
{
  bool ok;
  int ret = inputPeriodLe->text().toInt(&ok);
  if (ok && ret >= min_inp && ret <= max_inp) return ret;
  return orig_inputperiod;
}
int SettingsDialog::graphTime() const 
{ 
  bool ok;
  int ret = graphTimeLe->text().toInt();
  if (ok && ret >= min_gt && ret <= max_gt) return ret;
  return orig_graphtime;
}

void SettingsDialog::enableApply()
{
  apply->setEnabled(true);
}

void SettingsDialog::disableApply()
{
  apply->setEnabled(false);
}

void ClientMgr::doSettingsDialog()
{
  delete settingsDialog;
  settingsDialog = new SettingsDialog(inputPeriodHz, minInputPeriodHz, maxInputPeriodHz, graphTimeSecs, minGraphTimeSecs, maxGraphTimeSecs, 0, "ComediClient Settings Dialog", true);

  QObject::connect(settingsDialog, SIGNAL(applyClicked()), this, SLOT(settingsDialogApplyClicked()));

  if (settingsDialog->exec() == QDialog::Accepted) {
    inputPeriodHz = settingsDialog->inputPeriod();
    graphTimeSecs = settingsDialog->graphTime();    
  }
  delete settingsDialog;
  settingsDialog = 0;

  // do these unconditionally because apply may have temorarily broken these values..
  if (comm) 
    changeInput(form->chanAI->currentItem()); // re-creates data socket with new params
  applyGraphResize();
}

void ClientMgr::settingsDialogApplyClicked()
{
  if (settingsDialog) {
    int saved_inputperiod = inputPeriodHz;
    int saved_graphtime = graphTimeSecs;
    inputPeriodHz = settingsDialog->inputPeriod();
    graphTimeSecs = settingsDialog->graphTime();
    if (comm) changeInput(form->chanAI->currentItem()); // re-creates data socket with new params
    applyGraphResize();
    inputPeriodHz = saved_inputperiod; // revert back
    graphTimeSecs = saved_graphtime; // revert back
  }
}

void ClientMgr::doAboutDialog()
{
  QMessageBox::
    about(form, "About ComediClient", 
          QString("<p><font size=+1><B>") + form->caption() + "</b></font><br>"
          "A program for communicating with ComediServer.</p>"
          
          "<p>This program is licensed under the GPL v2 or above.</p>"
          
          "<p>Copyright (C) 2007 Calin A. Culianu &lt;calin@ajvar.org&gt;</p>");
}
