#ifndef ClientMgr_H
#define ClientMgr_H
#include <qobject.h>
#include <qsettings.h>
#include <qdns.h>
#include <qtimer.h>

#include "ClientUI.h"
#include "CID.h"
#include "ComediDevice.h"
#include <map>
#include <list>
#include <vector>
#include <set>

class QSocketDevice;
class QSocketNotifier;
class QLineEdit;
class SocketMgr;
class QwtPlot;
class QwtPlotCurve;
class QwtPicker;
class QPushButton;
class QCheckBox;
class QScrollView;
class DIOMgr;
class QScrollView;
class SettingsDialog;

struct CIDEx : public CID
{
  CIDEx() : CID(), isAllDIO(false) {}
  CIDEx(const CID & c) : CID(c), isAllDIO(false) {}

  bool isAllDIO : 1;
};

class ClientMgr : public QObject
{
  Q_OBJECT

public:
  ClientMgr(ClientUI *ui_parent, const char *name = 0);
  ~ClientMgr();

protected slots:
  void connect();
  void disconnect();
  void setConnected(bool);
  void gotSocketError(int);
  void sendFinished(int);
  void recvFinished(const QByteArray &);
  void pauseUnpause(bool);

  // slot called from the qlineedit in the form
  void setOutput();

  void doDIOConfigModify(CID, ComediSubDevice::DIOConfig);
  void doDIOWrite(CID, bool);

private:
  void doConnect();
  void connectError(int);

private slots:
  void resolvedHost();
  void resolverTimeout();
  
  bool changeInput(int); // param is index into form->chanAI
  bool changeInputGain(int); // param is index into form->gainAI
  void changeOutputGainBox(int); // param is index into form->chanAO

  void doKeepAlive(); // sends noop on protocol V2 or above..
  
  void doSettingsDialog();
  void doAboutDialog();
  void settingsDialogApplyClicked();

private:
  ClientUI *form;
  QSettings settings;
  ComediDevice *devices;
  unsigned numDevices;

  QString host, hostName;
  unsigned short port;

  enum SockState {
    Invalid = 0,
    ResolvingHost,
    SendingLine,
/*     ReceivingDeviceNums, */
/*     ReceivingDevices, */
/*     ReceivingDataKey, */
/*     SendingDataKey, */
/*     ReceivingDataKeyOK, */
    ReceivingData
  };
  
  typedef std::map<SocketMgr *, std::pair<SockState, SockState> > SockStateMap;
  SockStateMap sockStateMap;    
  QSocketDevice *comm;
  SocketMgr *data;
  typedef std::list<SocketMgr *> Datas;
  Datas datas;
  

  void doSocketMgrSignalConnections(SocketMgr *);
  void loadSettings();
  void saveSettings();
  bool getServerVersion(); // asks server for its version number -- some servers don't understand this command -- side effect is to set serverVersion.  serverVersion is set to 1 if the server doesn't uinderstand this commant
  bool getDevices(); // asks server for its list of comedi devices, sets state so asynch receive works
  bool gotDevices(); // called when server finishes transmitting the devices, updates UI
  void populateGains();
  void removeSocketMgr(SocketMgr *);
  bool createDataSocket(CIDEx c, unsigned gain = 0, unsigned hz = 0);
  bool startAllDIOACQs();
  void dioConfigSubdev(unsigned dev, unsigned subdev, 
                       ComediSubDevice::DIOConfig conf);
  void applyGraphResize();
  void createMenuBar();

  typedef std::vector<ComediSubDevice::DIOConfig> ConfigVec;
  ConfigVec dioGetConfig(unsigned dev, unsigned subdev);


  ComediDevice *findDeviceByMinor(unsigned minor) const;

  void gotDIOData(CIDEx cid, unsigned int chanBits);  

  QString receiveCommLine();
  bool sendCommLine(const QString &);
  static bool sendLine(QSocketDevice *s, const QString &);
  static QString receiveLine(QSocketDevice *s);

  bool doOkCmd(const QString &cmd, const QString &command_description);
  void saveAOAIComboBoxes();
  void restoreAOAIComboBoxes();

  typedef std::map<CIDEx, int> CIDCB;
  typedef std::map<int, CIDEx> CBCID;  
  
  CIDCB cidcb_in, cidcb_out;
  CBCID cbcid_in, cbcid_out;

  bool disconnecting, connecting;

  typedef std::map<CIDEx, SocketMgr *> CIDSM;
  typedef std::map<SocketMgr *, CIDEx> SMCID;
  CIDSM cidsm;
  SMCID smcid;

  SettingsDialog *settingsDialog;

  QwtPlot *graph; // the graph!
  QwtPlotCurve *curve; // the graph's curve
  QwtPicker *cursorPicker; // the graph's zoomer
  std::vector<double> graphDataX;
  std::vector<double> graphDataY;
  ComediRange *currentRange;
  double currentMaxData;
  bool graphPaused;
  QScrollView *dioSV;
  DIOMgr *dioMgr;
  typedef std::map<unsigned, ComediDevice *> MinorMap;
  MinorMap minorMap;
  QDns dns;
  QTimer dnsTimer, keepaliveTimer;
  QScrollView *scrollView;

  typedef std::set<CID> CIDSet;
  CIDSet exclusiveDIOs;
  
  unsigned serverVersion;

  struct SavedComboBoxIdxes {
    int out, outGain, in, inGain;
    SavedComboBoxIdxes() : out(0), outGain(0), in(0), inGain(0) {}
  };

  // remember state of last input/output channel for each host..
  std::map<QString, SavedComboBoxIdxes> savedCBStates;
  int inputPeriodHz, graphTimeSecs;
  static const int  dioInputPeriodHz, defInputPeriodHz, defGraphTimeSecs, minInputPeriodHz, maxInputPeriodHz, minGraphTimeSecs, maxGraphTimeSecs;
  static QString socketErrorToString(int);
};

class SettingsDialog : public QDialog
{
  Q_OBJECT

public:
  SettingsDialog(int inputperiod, int min_inp, int max_inp,
                 int graphtime, int min_graphtime, int max_graphtime,
                 QWidget *parent, const char *name = 0, bool modal = false);
  int inputPeriod() const;
  int graphTime() const;
signals:
  void applyClicked();
private slots:
  void enableApply();
  void disableApply();
private:
  int orig_inputperiod, orig_graphtime, min_inp, max_inp, min_gt, max_gt;
  QLineEdit *inputPeriodLe, *graphTimeLe;
  QPushButton *apply;
};

class DIOLed;

class DIOMgr : public QWidget
{
  Q_OBJECT
public:
  // note upon construction all DIO lines are considered to be set to INPUT mode
  DIOMgr(ComediDevice *devices, unsigned numDevs,
         QWidget *parent, const char *name = 0);
  ~DIOMgr();
    
public slots:
  void setDIOState(CID chan, bool highlow /* true is high */);
  void setDIOCB(CID chan, ComediSubDevice::DIOConfig); 
 
signals:
  void dioConfigModifyRequested(CID chan, ComediSubDevice::DIOConfig);
  void dioWriteRequested(CID chan, bool oneorzero);

private slots:
  void doDioToggle();
  void doDioChecked(bool);

private:
  typedef std::map<CID, DIOLed *> ButtonMap;
  typedef std::map<CID, QCheckBox *> CheckBoxMap;
  ButtonMap buttonMap;
  CheckBoxMap checkBoxMap;
  
};

class CIDRememberer
{
public:
  CIDRememberer() {}
  void setCID(CID c) { cid = c; }
  CID  getCID() const { return cid; }
private:
  CID cid;
};

class DIOLed : public QWidget, public CIDRememberer
{
  Q_OBJECT
public:
  DIOLed(QWidget *parent = 0, const char *name = 0);
  ~DIOLed();
  const QColor & highColor() const { return high; }
  const QColor & lowColor() const { return low; }
  bool state() const { return ishigh; }
  void setHighColor(const QColor & c) { high = c; repaint(); }
  void setLowColor(const QColor & c) { low = c; repaint(); }  
public slots:
  void setState(bool high);
protected:
  void paintEvent(QPaintEvent *);
  void mousePressEvent(QMouseEvent *);
signals:
  void clicked();
private:
  bool ishigh;
  QColor high, low;
};

#endif
