#include "SocketMgr.h"

SocketMgr::SocketMgr(Type t, QSocketDevice *s, QObject *parent, const char *name)
  : QObject(parent, name), 
    type(t),
    sock(s),
    readNot(sock->socket(), QSocketNotifier::Read, this),
    writeNot(sock->socket(), QSocketNotifier::Write, this)
{

  init();

  switch (type) {
  case Read:
    //setSocketRead(true);
    connect(&readNot, SIGNAL(activated(int)), this, SLOT(readReady()));
    break;
  case Write:
    //setSocketWrite(true);
    connect(&writeNot, SIGNAL(activated(int)), this, SLOT(writeReady()));
    break;
  }
}

SocketMgr::~SocketMgr()
{
  delete sock;
  sock = 0;
}

void SocketMgr::init()
{
  reset();
  sock->setBlocking(false);
}

void SocketMgr::reset()
{
  dataBuf.resize(0);
  numRecvDesired = 0;
  sent = 0;
  setSocketRead(false);
  setSocketWrite(false);
}

void SocketMgr::requestData(int numBytes)
{
  if (type != Read) return;
  numRecvDesired = numBytes;
  setSocketRead(true);
  checkReceiveIsDone();
}

void SocketMgr::enqueueData(const QByteArray &data)
{
  if (type != Write) return;
  unsigned oldsize = dataBuf.size();  
  dataBuf.resize(dataBuf.size() + data.size());
  unsigned i,j;
  // fill in data to dataBuf
  for (i = oldsize, j = 0; i < dataBuf.size(); ++i, ++j)
    dataBuf[i] = data[j];
  setSocketWrite(true);
}

void SocketMgr::readReady()
{
  unsigned nBytes = sock->bytesAvailable();
  int recvd = dataBuf.size();
  dataBuf.resize(recvd + nBytes);
  int ret = sock->readBlock(dataBuf.data() + recvd, nBytes);
  if (ret <= 0) {
    int err = sock->error();
    init();
    sock->close();
    emit socketError(err);
    return;
  }
  checkReceiveIsDone(); 
}

void SocketMgr::consumeData(unsigned numBytes)
{
  if (numBytes > dataBuf.size()) {
    numBytes = dataBuf.size();
  } 
  QByteArray copy;
  if (numBytes < dataBuf.size()) 
    copy.duplicate (dataBuf.data() + numBytes, dataBuf.size()-numBytes);
  dataBuf = copy;
}

void SocketMgr::checkReceiveIsDone()
{
  int recvd = dataBuf.size();
  if (numRecvDesired < 0) {
    // scan for newline, if we got it, emit gotLine()
    int pos;
    
    while ( (pos = dataBuf.find('\n'))  > -1) {
      // we got a newline!
      dataBuf[pos] = 0;      
      QString line(dataBuf.data());
      consumeData(pos+1);
      emit gotLine(line);
    }
  } else if (recvd >= numRecvDesired) {
    emit dataReceived(dataBuf);
    dataBuf.resize(0);
  }
}

void SocketMgr::writeReady()
{
    if (sent < dataBuf.size()) {
      int ret = sock->writeBlock(dataBuf.data() + sent, dataBuf.size()-sent);
      if (ret < 0) {
        int err = sock->error();
        emit socketError(err);
        init();
        sock->close();
        return;
      }
      sent += ret;
    }
    checkSendIsDone();
}

void SocketMgr::checkSendIsDone()
{
  if (sent >= dataBuf.size()) {
    setSocketWrite(false);
    emit dataSent(sent);
    reset();
  }
}

QString SocketMgr::socketErrorString(int e)
{
  switch(e) {
  case QSocketDevice::NoError: return "No error"; 
  case QSocketDevice::AlreadyBound: return "Already bound"; 
  case QSocketDevice::Inaccessible: return "Inaccessible"; 
  case QSocketDevice::NoResources: return "Out of resources";     
  case QSocketDevice::Impossible: return "Impossible";
  case QSocketDevice::ConnectionRefused: return "Connection refused";
  case QSocketDevice::NoFiles: return "Out of file descriptors";
  case QSocketDevice::NetworkFailure: return "Generic network failure";
  case QSocketDevice::InternalError: 
  default:
    return "Internal error"; 
  }
  return ""; // not reached
}

bool SocketMgr::isSocketRead() const { return readNot.isEnabled(); }
bool SocketMgr::isSocketWrite() const { return writeNot.isEnabled(); }

void SocketMgr::setSocketRead(bool b) { readNot.setEnabled(b); }
void SocketMgr::setSocketWrite(bool b) { writeNot.setEnabled(b); }
