#ifndef SockMgr_H
#define SockMgr_H

#include <qobject.h>
#include <qsocketdevice.h>
#include <qsocketnotifier.h>
#include <qcstring.h>

class SocketMgr : public QObject
{
  Q_OBJECT

public:
  enum Type { Read, Write };

  SocketMgr(Type t, QSocketDevice *thesocket, QObject *parent = 0, const char *name = 0);
  ~SocketMgr();
  QSocketDevice & socket() const { return *const_cast<QSocketDevice*>(sock); }

 /** Notify class that we want data in numBytes sized chunks.  This tells
     this class to emit dataReceived() signal when numBytes sized chunks
     are ready.

     Negative value means receive a full line (trailing \n is stripped).

     0/1 means emit dataReceived() as soon as *any* data is available, no
     matter how small.

     Positive value means emit dataReceived() signal when numBytes threshold 
     is met (that is, data received >= numBytes). 

     Only valid for a 'Read' SocketMgr.  */
  void requestData(int numBytes = -1);

  /** Transmit data in the QByteArray asynchronously.  Data is enqueued
      and sent in the main thread whenever the underlying socket notifies 
      that data can be consumed.   */
  void enqueueData(const QByteArray &);
  
  /** Resets the data buffer or any pending requests, etc. */
  void reset();
  
  static QString socketErrorString(int);

signals:
  // these two emitted when a send/receive cycle fully completes
  void dataSent(int);  // param is number of bytes sent
  void dataReceived(const QByteArray & data); // param is the data received 
  void gotLine(const QString &); // emitted only when numBytes to receive is negative -- meaning we are in line-by-line mode.

  void socketError(int); // emitted when we suspect the socket has closed, etc param is QSocketDevice::error() value

protected:
  void setSocketRead(bool);
  void setSocketWrite(bool);
  bool isSocketRead() const;
  bool isSocketWrite() const;

private slots:
  void readReady();
  void writeReady();

private:
  Type type;
  QSocketDevice *sock;
  QSocketNotifier readNot, writeNot;
  QByteArray dataBuf;
  unsigned sent;
  int numRecvDesired;

  void checkReceiveIsDone(); 
  void checkSendIsDone();
  void consumeData(unsigned numBytes);  
  void init();
};



#endif
