/* -*- mode: C++ ; c-file-style: "stroustrup" -*- *****************************
 * Qwt Widget Library
 * Copyright (C) 1997   Josef Wilgen
 * Copyright (C) 2002   Uwe Rathmann
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the Qwt License, Version 1.0
 *****************************************************************************/

#ifndef QWT_SCALE_DRAW_H
#define QWT_SCALE_DRAW_H

#include <qstring.h>
#include <qrect.h>
#include "qwt_global.h"
#include "qwt_scale_map.h"

class QPainter;
class QPen;
class QColorGroup;
class QPalette;
class QFontMetrics;
class QFont;
class QwtScaleDiv;

/*!
  \brief A class for drawing scales

  QwtScaleDraw can be used to draw linear or logarithmic scales.
  A scale has an origin,
  an orientation and a length, which all can be specified with
  QwtScaleDraw::setGeometry(). The labels can be rotated and aligned
  to the ticks using QwtScaleDraw::setLabelRotation(), 
  QwtScaleDraw::setLabelFlags().

  After a scale division has been specified as a QwtScaleDiv object
  using QwtScaleDraw::setScale(const QwtScaleDiv &s),
  or determined internally using QwtScaleDraw::setScale(double vmin, double vmax, int maxMajIntv, int maxMinIntv, double step = 0.0, int logarithmic = 0),
  the scale can be drawn with the QwtScaleDraw::draw() member.
*/
class QWT_EXPORT QwtScaleDraw
{
public:
    /*! 
        Orientation/Position of the scale draw
        \sa QwtScaleDraw::setGeometry
     */
    enum Orientation { Bottom, Top, Left, Right, Round };
    enum Option { None = 0, Backbone = 1 };
 
    QwtScaleDraw();
    QwtScaleDraw( const QwtScaleDraw & );
    virtual ~QwtScaleDraw();

    QwtScaleDraw &operator=(const QwtScaleDraw &);
    
    void setTransformation(const QwtScaleTransformation&);
    void setScaleDiv(const QwtScaleDiv &s);

    const QwtScaleMap &map() const;

    void setOptions(int opt);
    int options() const;

    void setGeometry(int xorigin, int yorigin, int length, Orientation o);
    int x() const;
    int y() const;
    int length() const;
    Orientation orientation() const;

    void setAngleRange(double angle1, double angle2);
    void setLabelFormat(char f, int prec, int fieldwidth = 0);
    void setLabelFlags(int flags);
    void setLabelRotation(double rotation);
    void setTickLength(unsigned int minLen, 
        unsigned int medLen, unsigned int majLen);

    void setMargin(uint hMargin, uint vMargin);

    const QwtScaleDiv& scaleDiv() const;
    int maxWidth(const QPen &, const QFont &) const;
    int maxHeight(const QPen &, const QFont &) const;

    void minBorderDist(const QFont &, int &start, int &end) const;
    int minLabelDist(const QFont &) const;

    int minHeight(const QPen &, const QFont &) const; 
    int minWidth(const QPen &, const QFont &) const; 
    void labelFormat(char &f, int &prec, int &fieldwidth) const;
    int labelFlags() const;
    double labelRotation() const;
    void tickLength(unsigned int &minLen, 
        unsigned int &medLen, unsigned int &majLen) const;
    unsigned int majTickLength() const;
        
#if QT_VERSION < 0x040000
    virtual void draw(QPainter *, const QColorGroup &) const;
#else
    virtual void draw(QPainter *, const QPalette &) const;
#endif
    virtual QString label(double) const;

protected:
    virtual void labelPlacement( const QFontMetrics &, double val,
        QPoint &pos, int &alignment, double &rotation) const;
    virtual int maxLabelHeight(const QFont &) const;
    virtual int maxLabelWidth(const QFont &) const;
    QRect labelBoundingRect(const QFont &, double val) const;

    void drawTick(QPainter *p, double val, int len) const;
    void drawBackbone(QPainter *p) const;
    void drawLabel(QPainter *p, double val) const;

#if QT_VERSION < 0x040000
    QWMatrix labelWorldMatrix(const QFontMetrics &,
        const QPoint &, int alignment, double rotation, 
        const QString &) const;
#else
    QMatrix labelWorldMatrix(const QFontMetrics &,
        const QPoint &, int alignment, double rotation, 
        const QString &) const;
#endif
        
private:
    int operator==(const QwtScaleDraw &) const;
    int operator!=(const QwtScaleDraw &) const;

    class PrivateData;
    PrivateData *d_data;
};

#endif
