/* -*- mode: C++ ; c-file-style: "stroustrup" -*- *****************************
 * Qwt Widget Library
 * Copyright (C) 1997   Josef Wilgen
 * Copyright (C) 2002   Uwe Rathmann
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the Qwt License, Version 1.0
 *****************************************************************************/

// vim: expandtab

#include <qpainter.h>
#include "qwt_painter.h"
#include "qwt_scale_map.h"
#include "qwt_plot_marker.h"
#include "qwt_symbol.h"
#include "qwt_text.h"
#include "qwt_math.h"

static const int LabelDist = 2;

class QwtPlotMarker::PrivateData
{
public:
    PrivateData():
        align(Qt::AlignCenter),
        style(NoLine),
        xValue(0.0),
        yValue(0.0)
    {
        // QwtPlotMarker.draw() assumes that 
        // label->alignment() == Qt::AlignCenter. 

        label = QwtText::makeText(QString::null, Qt::AlignCenter, QFont());
    }

    ~PrivateData()
    {
        delete label;
    }

    QwtText *label;
    int align;
    QPen pen;
    QwtSymbol sym;
    LineStyle style;

    double xValue;
    double yValue;
};

//! Sets alignment to Qt::AlignCenter, and style to NoLine
QwtPlotMarker::QwtPlotMarker()
{
    d_data = new PrivateData;
    setZ(30.0);
}

//! Destructor
QwtPlotMarker::~QwtPlotMarker()
{
    delete d_data;
}

/*!
  \brief Copy constructor
  \param m Marker
*/
QwtPlotMarker::QwtPlotMarker(const QwtPlotMarker &m):
    QwtPlotItem(m)
{
    *this = m;
}

/*!
  \brief Assignment operator
  \param m Marker
*/
QwtPlotMarker& QwtPlotMarker::operator=(const QwtPlotMarker &m)
{
    if (this != &m)
    {
        QwtPlotItem::operator=((const QwtPlotItem &)m);

        d_data->label = m.d_data->label->clone();
        d_data->align = m.d_data->align;
        d_data->pen = m.d_data->pen;
        d_data->sym = m.d_data->sym;
        d_data->style = m.d_data->style;
        d_data->xValue = m.d_data->xValue;
        d_data->yValue = m.d_data->yValue;
    }

    return *this;
}

int QwtPlotMarker::rtti() const
{
    return QwtPlotItem::Rtti_PlotMarker;
}

QwtDoublePoint QwtPlotMarker::value() const
{
    return QwtDoublePoint(d_data->xValue, d_data->yValue);
}

//! Return x Value
double QwtPlotMarker::xValue() const 
{ 
    return d_data->xValue; 
}

//! Return y Value
double QwtPlotMarker::yValue() const 
{ 
    return d_data->yValue; 
}

void QwtPlotMarker::setValue(const QwtDoublePoint& pos)
{
    setValue(pos.x(), pos.y());
}

void QwtPlotMarker::setValue(double x, double y) 
{
    if ( x != d_data->xValue || y != d_data->yValue )
    {
        d_data->xValue = x; 
        d_data->yValue = y; 
        itemChanged(); 
    }
}

//! Set X Value
void QwtPlotMarker::setXValue(double x) 
{ 
    setValue(x, d_data->yValue);
}

//! Set Y Value
void QwtPlotMarker::setYValue(double y) 
{ 
    setValue(d_data->xValue, y);
}

/*!
  \brief Draw the marker
  \param p Painter
*/
void QwtPlotMarker::draw(QPainter *p,
    const QwtScaleMap &xMap, const QwtScaleMap &yMap,
    const QRect &r) const
{
    const int x = xMap.transform(d_data->xValue);
    const int y = yMap.transform(d_data->yValue);

    // QwtPlotMarker.draw() assumes that 
    // d_data->label->alignment() == Qt::AlignCenter. 

    // draw lines
    if (d_data->style != NoLine)
    {
        p->setPen(d_data->pen);
        if ((d_data->style == HLine) || (d_data->style == Cross))
            QwtPainter::drawLine(p, r.left(), y, r.right(), y);
        if ((d_data->style == VLine)||(d_data->style == Cross))
            QwtPainter::drawLine(p, x, r.top(), x, r.bottom());
    }

    // draw symbol
    QSize sSym(0, 0);
    if (d_data->sym.style() != QwtSymbol::None)
    {
        sSym = d_data->sym.size();
        d_data->sym.draw(p, x, y);
    }

    // draw label
    if (!d_data->label->text().isEmpty())
    {
        int xlw = qwtMax(int(d_data->pen.width()), 1);
        int ylw = xlw;
        int xlw1;
        int ylw1;

        const int xLabelDist = 
            QwtPainter::metricsMap().screenToLayoutX(LabelDist);
        const int yLabelDist = 
            QwtPainter::metricsMap().screenToLayoutY(LabelDist);

        if ((d_data->style == VLine) || (d_data->style == HLine))
        {
            xlw1 = (xlw + 1) / 2 + xLabelDist;
            xlw = xlw / 2 + xLabelDist;
            ylw1 = (ylw + 1) / 2 + yLabelDist;
            ylw = ylw / 2 + yLabelDist;
        }
        else 
        {
            xlw1 = qwtMax((xlw + 1) / 2, (sSym.width() + 1) / 2) + xLabelDist;
            xlw = qwtMax(xlw / 2, (sSym.width() + 1) / 2) + xLabelDist;
            ylw1 = qwtMax((ylw + 1) / 2, (sSym.height() + 1) / 2) + yLabelDist;
            ylw = qwtMax(ylw / 2, (sSym. height() + 1) / 2) + yLabelDist;
        }

        // tr is offset with respect to (0, 0) obeying the immutable flags 
        // Qt::AlignCenter. Hence: 
        // tr.x() == -tr.width()/2
        // tr.y() == -tr.height()/2
        QRect tr = d_data->label->boundingRect(p);

        int dx = x;
        int dy = y;

        if (d_data->style == VLine)
        {
            if (d_data->align & (int) Qt::AlignTop)
                dy = r.top() + yLabelDist - tr.y();
            else if (d_data->align & (int) Qt::AlignBottom)
                dy = r.bottom() - yLabelDist + tr.y();
            else
                dy = r.top() + r.height() / 2;
        }
        else
        {
            if (d_data->align & (int) Qt::AlignTop)
                dy += tr.y() - ylw1;
            else if (d_data->align & (int) Qt::AlignBottom)
                dy -= tr.y() - ylw1;
        }


        if (d_data->style == HLine)
        {
            if (d_data->align & (int) Qt::AlignLeft)
                dx = r.left() + xLabelDist - tr.x();
            else if (d_data->align & (int) Qt::AlignRight)
                dx = r.right() - xLabelDist + tr.x();
            else
                dx = r.left() + r.width() / 2;
        }
        else
        {
            if (d_data->align & (int) Qt::AlignLeft)
                dx += tr.x() - xlw1;
            else if (d_data->align & (int) Qt::AlignRight)
                dx -= tr.x() - xlw1;
        }

#if QT_VERSION < 0x040000
        tr.moveBy(dx, dy);
#else
        tr.translate(dx, dy);
#endif
        d_data->label->draw(p, tr);
    }
}

/*!
  \brief Specify the font for the label
  \param f New font
  \sa QwtPlotMarker::font()
*/
void QwtPlotMarker::setFont(const QFont &f)
{
    if ( f == d_data->label->font() )
        return;
    
    d_data->label->setFont(f);
    itemChanged();
}

/*!
  \return the font used for the label
  \sa QwtPlotMarker::setFont()
*/
const QFont QwtPlotMarker::font() const 
{ 
    return d_data->label->font(); 
}


/*!
  \brief Set the line style
  \param st Line style. Can be one of QwtPlotMarker::NoLine,
    QwtPlotMarker::HLine, QwtPlotMarker::VLine or QwtPlotMarker::Cross
  \sa QwtPlotMarker::lineStyle()
*/
void QwtPlotMarker::setLineStyle(QwtPlotMarker::LineStyle st)
{
    if ( st != d_data->style )
    {
        d_data->style = st;
        itemChanged();
    }
}

/*!
  \return the line style
  \sa For a description of line styles, see QwtPlotMarker::setLineStyle()
*/
QwtPlotMarker::LineStyle QwtPlotMarker::lineStyle() const 
{ 
    return d_data->style; 
}

/*!
  \brief Assign a symbol
  \param s New symbol 
  \sa QwtSymbol, QwtPlotMarker::symbol()
*/
void QwtPlotMarker::setSymbol(const QwtSymbol &s)
{
    d_data->sym = s;
    itemChanged();
}

/*!
  \return the symbol
  \sa QwtPlotMarker::setSymbol(), QwtSymbol
*/
const QwtSymbol &QwtPlotMarker::symbol() const 
{ 
    return d_data->sym; 
}

/*!
  \brief Set the label text
  \param text label text 
  \sa QwtPlotMarker::label()
*/
void QwtPlotMarker::setLabelText(const QString &text)
{
    setLabel(text, d_data->label->font(), d_data->label->color(), 
        d_data->label->rectPen(), d_data->label->rectBrush());
}

/*!
  \brief Set the label
  \param text Contents of the label text
  \param font Font of the label text
  \param color Color of the label text
  \param pen Pen to draw the enclosing rectangle
  \param brush Brush to draw the enclosing rectangle
  \sa QwtPlotMarker::label()
*/
void QwtPlotMarker::setLabel(const QString &text, const QFont &font,
    const QColor &color, const QPen &pen, const QBrush &brush)
{
    if ( text == d_data->label->text()
         && font == d_data->label->font()
         && color == d_data->label->color()
         && pen == d_data->label->rectPen()
         && brush == d_data->label->rectBrush() )
        return;
    
    QwtText *label = QwtText::makeText(
        text, d_data->label->flags(), font, color, pen, brush);

    delete d_data->label;
    d_data->label = label;

    itemChanged();
}

/*!
  \return the label
  \sa QwtPlotMarker::setLabel()
*/
const QString QwtPlotMarker::label() const 
{ 
    return d_data->label->text(); 
}

/*!
  \brief Set the alignment of the label

  The alignment determines where the label is drawn relative to
  the marker's position.

  \param align Alignment. A combination of AlignTop, AlignBottom,
    AlignLeft, AlignRight, AlignCenter, AlgnHCenter,
    AlignVCenter.  
  \sa QwtPlotMarker::labelAlignment()
*/
void QwtPlotMarker::setLabelAlignment(int align)
{
    if ( align == d_data->align )
        return;
    
    d_data->align = align;
    itemChanged();
}

/*!
  \return the label alignment
  \sa QwtPlotMarker::setLabelAlignment()
*/
int QwtPlotMarker::labelAlignment() const 
{ 
    return d_data->align; 
}

/*!
  \brief Specify a pen for the line.
  \param p New pen
  \sa QwtPlotMarker::linePen()
*/
void QwtPlotMarker::setLinePen(const QPen &p)
{
    if ( p != d_data->pen )
    {
        d_data->pen = p;
        itemChanged();
    }
}

/*!
  \return the line pen
  \sa QwtPlotMarker::setLinePen()
*/
const QPen &QwtPlotMarker::linePen() const 
{ 
    return d_data->pen; 
}

/*!
  \brief Specify a color for the label
  \param color Color of the label text
  \sa QwtPlotMarker::labelColor()
*/
void QwtPlotMarker::setLabelColor(const QColor &color)
{
    if ( color == d_data->label->color() )
        return;
    
    d_data->label->setColor(color);
    itemChanged();
}

/*!
  \return the label's color
  \sa QwtPlotMarker::setLabelPen()
*/
const QColor QwtPlotMarker::labelColor() const 
{ 
    return d_data->label->color(); 
}

/*!
  \brief Specify a pen for the label
  \param p New pen
  \sa QwtPlotMarker::labelPen()
  \warning This function is deprecated. Use QwtPlotMarker::setLabelColor.
*/
void QwtPlotMarker::setLabelPen(const QPen &p)
{
    if ( p == QPen(d_data->label->color()) )
        return;
    
    d_data->label->setColor(p.color());
    itemChanged();
}

/*!
  \return the label's pen
  \sa QwtPlotMarker::setLabelPen()
  \warning This function is deprecated. Use QwtPlotMarker::labelColor.
*/
const QPen QwtPlotMarker::labelPen() const 
{ 
    return QPen(d_data->label->color()); 
}

QwtDoubleRect QwtPlotMarker::boundingRect() const
{
    return QwtDoubleRect(d_data->xValue, d_data->yValue, 0.0, 0.0);
}
