/* -*- mode: C++ ; c-file-style: "stroustrup" -*- *****************************
 * Qwt Widget Library
 * Copyright (C) 1997   Josef Wilgen
 * Copyright (C) 2002   Uwe Rathmann
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the Qwt License, Version 1.0
 *****************************************************************************/

// vim: expandtab

#include <stdio.h>
#include <qpainter.h>
#include <qpalette.h>
#include <qwidget.h>
#include "qwt_math.h"
#include "qwt_painter.h"
#include "qwt_layout_metrics.h"
#include "qwt_scale_div.h"
#include "qwt_scale_draw.h"
#if QT_VERSION < 0x040000
#include <qwmatrix.h>
#define QwtMatrix QWMatrix
#else
#include <qmatrix.h>
#define QwtMatrix QMatrix
#endif

class QwtScaleDraw::PrivateData
{
public:
    PrivateData():
        options(QwtScaleDraw::Backbone),
        hpad(4),
        vpad(4),
        medLen(6),
        majLen(8),
        minLen(4),
        minAngle(-135 * 16),
        maxAngle(135 * 16),
        fmt('g'),
        prec(4),
        fieldwidth(0),
        labelFlags(0),
        labelRotation(0.0)
    {
        // snprintf is C99 and therefore not portable :-(
        // fieldwidth and precision must be in the range 0, 1, .., 99.
        sprintf(formatBuffer, "%%%d.%d%c", fieldwidth, prec, fmt);
    }

    int options;
    
    QwtScaleMap map;
    QwtScaleDiv scldiv;
    QwtScaleDraw::Orientation orient;
        
    int xorg;
    int yorg;
    int len;
        
    unsigned int hpad;
    unsigned int vpad;

    unsigned int medLen;
    unsigned int majLen;
    unsigned int minLen;

    int minAngle;
    int maxAngle;

    double xCenter;
    double yCenter;
    double radius;

    char fmt;
    int prec;
    int fieldwidth;

    int labelFlags;
    double labelRotation;

    // formatBuffer must be able to hold:
    // - a percent sign a '%'
    // - a two digit positive fieldwidth
    // - a point
    // - a two digit positive precision
    // - a format character ('e', 'f', 'g', 'E', 'F, 'G')
    // - a terminating '/0' character
    char formatBuffer[8];
};

/*!
  \brief Constructor

  The range of the scale is initialized to [0, 100],
  the angle range is set to [-135, 135], the geometry
  is initialized such that the origin is at (0,0), the
  length is 100, and the orientation is QwtScaleDraw::Bottom.
*/
QwtScaleDraw::QwtScaleDraw()
{
    d_data = new QwtScaleDraw::PrivateData;
    setGeometry(0,0,100,Bottom);
}

QwtScaleDraw::QwtScaleDraw(const QwtScaleDraw &other)
{
    d_data = new QwtScaleDraw::PrivateData(*other.d_data);
}


//! Destructor
QwtScaleDraw::~QwtScaleDraw()
{
    delete d_data;
}

QwtScaleDraw &QwtScaleDraw::operator=(const QwtScaleDraw &other)
{
    *d_data = *other.d_data;
    return *this;
}

//! Change the scale draw options
void QwtScaleDraw::setOptions(int opt)
{
    d_data->options = opt;
}

/*!
  \return scale draw options
*/
int QwtScaleDraw::options() const
{
    return d_data->options;
}

/*!
  \brief Change the scale division
  \param sd new scale division
*/
void QwtScaleDraw::setScaleDiv(const QwtScaleDiv &sd)
{
    d_data->scldiv = sd;
    d_data->map.setScaleInterval(sd.lBound(), sd.hBound());
}

void QwtScaleDraw::setTransformation(
    const QwtScaleTransformation &transformation)
{
    d_data->map.setTransformation(transformation);
}

const QwtScaleMap &QwtScaleDraw::map() const
{
    return d_data->map;
}

/*! \return scale division */
const QwtScaleDiv& QwtScaleDraw::scaleDiv() const 
{ 
    return d_data->scldiv; 
}

/*!
  \brief Draw the scale

  \param painter    The painter
#if QT_VERSION < 0x040000
  \param colorGroup Color group, text color is used for the labels, 
                    foreground color for ticks and backbone
#else
  \param palette    Palette, text color is used for the labels, 
                    foreground color for ticks and backbone
#endif
*/
#if QT_VERSION < 0x040000
void QwtScaleDraw::draw(QPainter *painter, const QColorGroup& colorGroup) const
#else
void QwtScaleDraw::draw(QPainter *painter, const QPalette& palette) const
#endif
{
    uint i;

    QPen pen = painter->pen();

#if QT_VERSION < 0x040000
    painter->setPen(colorGroup.text()); // ignore pen style
#else
    painter->setPen(palette.color(QPalette::Text)); // ignore pen style
#endif

    const QwtTickList &majorTicks = 
        d_data->scldiv.ticks(QwtScaleDiv::MajorTick);

    for (i=0; i< (uint)majorTicks.count(); i++)
    {
        const double v = majorTicks[i];
        if ( d_data->scldiv.contains(v) )
            drawLabel(painter, majorTicks[i]);
    }

#if QT_VERSION < 0x040000
    pen.setColor(colorGroup.foreground()); // respect pen style
#else
    pen.setColor(palette.color(QPalette::Foreground)); // respect pen style
#endif

    int tickLength[QwtScaleDiv::NTickTypes];
    tickLength[QwtScaleDiv::MinorTick] = d_data->minLen;
    tickLength[QwtScaleDiv::MediumTick] = d_data->medLen;
    tickLength[QwtScaleDiv::MajorTick] = d_data->majLen;

    for ( int tickType = QwtScaleDiv::MinorTick; 
        tickType < QwtScaleDiv::NTickTypes; tickType++ )
    {
        const QwtTickList &ticks = d_data->scldiv.ticks(tickType);
        for (i=0; i < (uint)ticks.count(); i++)
        {
            const double v = ticks[i];
            if ( d_data->scldiv.contains(v) )
                drawTick(painter, ticks[i], tickLength[tickType]);
        }
    }

    if ( options() & Backbone )
        drawBackbone(painter);
}


//! Draws a single tick
void QwtScaleDraw::drawTick(QPainter *painter, double val, int len) const
{
    if ( len <= 0 )
        return;

    const int tval = map().transform(val);

    switch(d_data->orient)
    {
        case Left:
            QwtPainter::drawLine(painter, d_data->xorg, tval, 
                d_data->xorg - len, tval);
            break;

        case Right:
            QwtPainter::drawLine(painter, d_data->xorg, tval, 
                d_data->xorg + len, tval);
            break;

        case Bottom: 
            QwtPainter::drawLine(painter, tval, d_data->yorg, 
                tval, d_data->yorg + len);
            break;

        case Top:
            QwtPainter::drawLine(painter, tval, d_data->yorg, 
                tval, d_data->yorg - len);
            break;

        case Round:
            if ((tval <= d_data->minAngle + 359 * 16) 
                || (tval >= d_data->minAngle - 359 * 16))
            {
                const double arc = double(tval) / 16.0 * M_PI / 180.0;
                const int x1 = qRound( d_data->xCenter + sin(arc) 
                    * d_data->radius);
                const int x2 = qRound(d_data->xCenter + sin(arc) 
                    * (d_data->radius + double(len)));
                const int y1 = qRound(d_data->yCenter - cos(arc) 
                    * d_data->radius);
                const int y2 = qRound(d_data->yCenter - cos(arc) 
                    * (d_data->radius + double(len)));

                QwtPainter::drawLine(painter, x1, y1, x2, y2);
            }
            break;
    }
}

//! Draws the number label for a major scale tick
void QwtScaleDraw::drawLabel(QPainter *painter, double val) const
{
    QPoint pos;
    int flags;
    double rotation;
    labelPlacement(QFontMetrics(painter->font()), val, pos, flags, rotation);

    if ( flags )
    {
        const QString txt = label(val);
        if ( !txt.isEmpty() )
        {
            QwtMatrix m = labelWorldMatrix(QFontMetrics(painter->font()),
                pos, flags, rotation, txt);

            painter->save();
#ifndef QT_NO_TRANSFORMATIONS
#if QT_VERSION < 0x040000
            painter->setWorldMatrix(m, true);
#else
            painter->setMatrix(m, true);
#endif
#else
            painter->translate(m.dx(), m.dy());
#endif
            QwtPainter::drawText(painter, 0, 0, txt);
            painter->restore();
        }
    }
}

//! Find position, flags and rotation of the label
void QwtScaleDraw::labelPlacement( const QFontMetrics &fm, double val, 
    QPoint &pos, int &flags, double &rotation) const
{
    const int tval = map().transform(val);

    int x = 0;
    int y = 0;
    int flgs = 0;

    switch(d_data->orient)
    {
        case Right:
        {
            x = d_data->xorg + d_data->majLen + d_data->hpad + 1;
            y = tval;
            flgs = d_data->labelFlags;
            if ( flgs == 0 )
                flgs = Qt::AlignRight | Qt::AlignVCenter;
            break;
        }
        case Left:
        {
            x = d_data->xorg - d_data->majLen - d_data->hpad - 1;
            y = tval;
            flgs = d_data->labelFlags;
            if ( flgs == 0 )
                flgs = Qt::AlignLeft | Qt::AlignVCenter;
            break;
        }
        case Bottom:
        {
            x = tval;
            y = d_data->yorg + d_data->majLen + d_data->vpad + 1;
            flgs = d_data->labelFlags;
            if ( flgs == 0 )
                flgs = Qt::AlignHCenter | Qt::AlignBottom;
            break;
        }
        case Top:
        {
            x = tval;
            y = d_data->yorg - d_data->majLen - d_data->vpad - 1;
            flgs = d_data->labelFlags;
            if ( flgs == 0 )
                flgs = Qt::AlignHCenter | Qt::AlignTop;
            break;
        }
        case Round:
        {
            if ((tval > d_data->minAngle + 359 * 16) 
                || (tval < d_data->minAngle - 359 * 16))
            {
               break;
            }
            
            const int fmh = fm.ascent() - 2; 
            const double arc = tval / 16.0 / 360.0 * 2 * M_PI;
            const double radius = d_data->radius + d_data->majLen + d_data->vpad;

            // First we find the point on a circle enlarged
            // by half of the font height.

            double xOffset = ( radius + fmh / 2 ) * sin(arc); 
            double yOffset = ( radius + fmh / 2 ) * cos(arc);

            if ( qRound(xOffset) != 0 ) 
            {
                // The centered label might cut the circle 
                // with distance: d_data->radius + d_data->majLen + d_data->vpad
                // We align the label to the circle by moving
                // the x-coordinate, because we have only
                // horizontal labels here.

                const int fmw = fm.width(label(val));

                const double circleX = radius * sin(arc); 
                if ( xOffset < 0 )
                    xOffset = circleX - fmw / 2; // left
                else
                    xOffset = circleX + fmw / 2; // right
            }
            x = qRound(d_data->xCenter + xOffset);
            y = qRound(d_data->yCenter - yOffset);
            flgs = Qt::AlignHCenter | Qt::AlignVCenter;

            break;
        }
    }

    pos = QPoint(x, y);
    flags = flgs;
    rotation = d_data->labelRotation;
}

//! Return the world matrix for painting the label 
QwtMatrix QwtScaleDraw::labelWorldMatrix(const QFontMetrics &fm,
    const QPoint &pos, int flags, 
#ifdef QT_NO_TRANSFORMATIONS
    double,
#else
    double rotation, 
#endif
    const QString &txt) const
{
    const int w = fm.boundingRect(0, 0, 
        QWIDGETSIZE_MAX, QWIDGETSIZE_MAX, 0, txt).width() - 2;
    const int h = fm.ascent() - 2;
    
    int x, y;
    if ( flags & Qt::AlignLeft )
        x = -w;
    else if ( flags & Qt::AlignRight )
        x = 0 - w % 2;
    else // Qt::AlignHCenter
        x = -(w / 2);
        
    if ( flags & Qt::AlignTop )
        y = 0;
    else if ( flags & Qt::AlignBottom )
        y = h - 1;
    else // Qt::AlignVCenter
        y = h / 2;
    
    QwtMatrix m;
    m.translate(pos.x(), pos.y());
#ifndef QT_NO_TRANSFORMATIONS
    m.rotate(rotation);
#endif
    m.translate(x, y);

    return m;
}

//! Draws the baseline of the scale
void QwtScaleDraw::drawBackbone(QPainter *painter) const
{
    const int bw2 = painter->pen().width() / 2;
    
    switch(d_data->orient)
    {
        case Left:
            QwtPainter::drawLine(painter, d_data->xorg - bw2, 
                d_data->yorg, d_data->xorg - bw2, d_data->yorg + d_data->len - 1);
            break;
        case Right:
            QwtPainter::drawLine(painter, d_data->xorg + bw2, 
                d_data->yorg, d_data->xorg + bw2, d_data->yorg + d_data->len - 1);
            break;
        case Top:
            QwtPainter::drawLine(painter, d_data->xorg, d_data->yorg - bw2, 
                d_data->xorg + d_data->len - 1, d_data->yorg - bw2);
            break;
        case Bottom:
            QwtPainter::drawLine(painter, d_data->xorg, d_data->yorg + bw2, 
                d_data->xorg + d_data->len - 1, d_data->yorg + bw2);
            break;
        case Round:
        {
            const int a1 = qRound(qwtMin(map().p1(), map().p2()) - 90 * 16); 
            const int a2 = qRound(qwtMax(map().p1(), map().p2()) - 90 * 16); 
            
            painter->drawArc(d_data->xorg, d_data->yorg, d_data->len, 
                d_data->len, -a2, a2 - a1 + 1);           // counterclockwise
            break;
        }
    }
}


/*!
  \brief Specify the geometry of the scale

  The parameters xorigin, yorigin and length have different meanings,
  dependent on the
  orientation:
  <dl>
  <dt>QwtScaleDraw::Left
  <dd>The origin is the topmost point of the
      baseline. The baseline is a vertical line with the
      specified length. Scale marks and labels are drawn
      at the left of the baseline.
  <dt>QwtScaleDraw::Right
  <dd>The origin is the topmost point of the
      baseline. The baseline is a vertical line with the
      specified length. Scale marks and labels are drawn
      at the right of the baseline.
  <dt>QwtScaleDraw::Top
  <dd>The origin is the leftmost point of the
      baseline. The baseline is a horizontal line with the
      specified length. Scale marks and labels are drawn
      above the baseline.
  <dt>QwtScaleDraw::Bottom
  <dd>The origin is the leftmost point of the
      baseline. The baseline is a horizontal line with the
      specified length. Scale marks and labels are drawn
      below the baseline.
  <dt>QwtScaleDraw::Round
  <dd>The origin is the top left corner of the
      bounding rectangle of the baseline circle. The baseline
      is the segment of a circle with a diameter of the specified length.
      Scale marks and labels are drawn outside the baseline
      circle.</dl>

  \param xorigin x coordinate of the origin
  \param yorigin y coordinate of the origin
  \param length length or diameter of the scale, excluding border distance
  \param o The orientation
*/
void QwtScaleDraw::setGeometry(int xorigin, int yorigin, 
    int length, Orientation o)
{
    static int minLen = 10;

    d_data->xorg = xorigin;
    d_data->yorg = yorigin;
    d_data->radius = double(length) * 0.5;
    d_data->xCenter = double(xorigin) + double(length) * 0.5;
    d_data->yCenter = double(yorigin) + double(length) * 0.5;
    
    if (length > minLen)
       d_data->len = length;
    else
       d_data->len = minLen;
    
    d_data->orient = o;
    
    switch(d_data->orient)
    {
        case Left:
        case Right:
            d_data->map.setPaintInterval(d_data->yorg + d_data->len - 1, d_data->yorg);
            break;
        case Round:
            d_data->map.setPaintInterval(d_data->minAngle, d_data->maxAngle);
            break;
        case Top:
        case Bottom:
            d_data->map.setPaintInterval(d_data->xorg, d_data->xorg + d_data->len - 1);
            break;
    }
}

/*!
  \param pen pen
  \param font Font
  \return the maximum width of the scale 
*/
int QwtScaleDraw::maxWidth(const QPen &pen, const QFont &font) const
{
    int w = 0;

    switch (d_data->orient)
    {
        case Left:
        case Right:
            w += pen.width() + d_data->majLen 
                    + d_data->hpad + maxLabelWidth(font);
            break;
        case Round:
            w += pen.width() + d_data->majLen 
                    + d_data->vpad + maxLabelWidth(font);
            break;
        case Top:
        case Bottom:
            w = d_data->len + maxLabelWidth(font);
            break;
    }
    return w;
}

/*!
  \param pen pen
  \param font Font
  \return the maximum height of the scale
*/
int QwtScaleDraw::maxHeight(const QPen &pen, const QFont &font) const 
{
    int h = 0;
    
    switch (d_data->orient)
    {
        case Top:
        case Bottom:
            h = pen.width() + d_data->vpad 
                    + d_data->majLen + maxLabelHeight(font);
            break;
        case Left:
        case Right:
            h = d_data->len + maxLabelHeight(font);
            break;
        case Round:
            h = d_data->vpad + d_data->majLen;
            if ( maxLabelHeight(font) > 0 )
            {
                const QFontMetrics fm(font);
                h += fm.ascent() - 2;
            }
            break;
    }
    
    return h;
}

/*!
  Rotate all labels. 

  \param rotation Angle in degrees. When changing the label rotation,
                  the label flags might be adjusted too.
  \sa setLabelAlignment(), labelRotation(), labelAlignment().
  \warning Rotation of labels is not implemented for round scales.

*/
void QwtScaleDraw::setLabelRotation(double rotation)
{
    d_data->labelRotation = rotation;
}

/*!
  \return the label rotation
  \sa setLabelRotation(), labelAlignment()
*/
double QwtScaleDraw::labelRotation() const
{
    return d_data->labelRotation;
}

/*!
  Labels are aligned to the point ticklength + margin away from the backbone. 
  The alignment is relative to the orientation of the label text.
  In case of an flags of 0 the label will be aligned 
  depending on the orientation of the scale:

      QwtScaleDraw::Top: Qt::AlignHCenter | Qt::AlignTop\n
      QwtScaleDraw::Bottom: Qt::AlignHCenter | Qt::AlignBottom\n
      QwtScaleDraw::Left: Qt::AlignLeft | Qt::AlignVCenter\n
      QwtScaleDraw::Right: Qt::AlignRight | Qt::AlignVCenter\n

  Changing the alignment is often necessary for rotated labels.

  \param flags Or'd Qt::AlignmentFlags, Qt::TextFlags <see qnamespace.h>
  \warning Alignment of labels is not implemented for round scales.
  \sa QwtScaleDraw::setLabelRotation(), QwtScaleDraw::labelRotation(), 
      QwtScaleDraw::labelAlignment()
*/
  
void QwtScaleDraw::setLabelFlags(int flags)
{
    d_data->labelFlags = flags;
}

/*!
  \return the label flags
  \sa setLabelFlags(), labelRotation()
*/
int QwtScaleDraw::labelFlags() const
{
    return d_data->labelFlags;
}

/*!
  \brief Adjust the baseline circle segment for round scales.

  The baseline will be drawn from min(angle1,angle2) to max(angle1, angle2).
  The settings have no effect if the scale orientation is not set to
  QwtScaleDraw::Round. The default setting is [ -135, 135 ].
  An angle of 0 degrees corresponds to the 12 o'clock position,
  and positive angles count in a clockwise direction.
  \param angle1
  \param angle2 boundaries of the angle interval in degrees.
  \warning <ul>
  <li>The angle range is limited to [-360, 360] degrees. Angles exceeding
      this range will be clipped.
  <li>For angles more than 359 degrees above or below min(angle1, angle2),
      scale marks will not be drawn.
  <li>If you need a counterclockwise scale, use QwtScaleDiv::setRange
  </ul>
*/
void QwtScaleDraw::setAngleRange(double angle1, double angle2)
{
    angle1 = qwtLim(angle1, -360.0, 360.0);
    angle2 = qwtLim(angle2, -360.0, 360.0);

    int amin = qRound(qwtMin(angle1, angle2) * 16.0);
    int amax = qRound(qwtMax(angle1, angle2) * 16.0);
    
    if (amin == amax)
    {
        amin -= 1;
        amax += 1;
    }
    
    d_data->minAngle = amin;
    d_data->maxAngle = amax;
    d_data->map.setPaintInterval(d_data->minAngle, d_data->maxAngle);
}

/*!
  \brief Set the number format for the major scale labels

  Format character, precision and fieldwidth have the same meaning as for
  sprintf().
  \param f format character 'e', 'f', 'g' 
  \param prec
    - for 'e', 'f': the number of digits after the radix character (point)
    - for 'g': the maximum number of significant digits
  \param fieldwidth fieldwidth

  \note precision and fieldwidth must be in the range 0, 1, .., 99.  Invalid
  input is discarded.

  \sa labelFormat()
*/
void QwtScaleDraw::setLabelFormat(char f, int prec, int fieldwidth)
{
    switch(f)
    {
        case 'e':
        case 'f':
        case 'g':
            d_data->fmt = f;
            break;
        default:
            break;
    }
    if ((prec > -1) && (prec < 99))
        d_data->prec = prec;
    if ((fieldwidth > -1) && (fieldwidth < 99))
        d_data->fieldwidth = fieldwidth;

    sprintf(d_data->formatBuffer, "%%%d.%d%c", d_data->fieldwidth, d_data->prec, d_data->fmt);
}

/*!
  \brief Return the number format for the major scale labels

  Format character, precision and fieldwidth have the same meaning as for
  sprintf().
  \param f format character 'e', 'f' or 'g' 
  \param prec
    - for 'e', 'f': the number of digits after the radix character (point)
    - for 'g': the maximum number of significant digits
  \param fieldwidth fieldwidth

  \sa setLabelFormat()
*/
void QwtScaleDraw::labelFormat(char &f, int &prec, int &fieldwidth) const
{
    f = d_data->fmt;
    prec = d_data->prec;
    fieldwidth = d_data->fieldwidth;
}

/*!
    Set the margins of the ticks

    \param hMargin Horizontal margin
    \param vMargin Vertical margin
*/
void QwtScaleDraw::setMargin(uint hMargin, uint vMargin)
{
    d_data->hpad = hMargin;
    d_data->vpad = vMargin;
}

/*!
    Set the length of the ticks
*/
void QwtScaleDraw::setTickLength(unsigned int minLen, 
    unsigned int medLen, unsigned int majLen)
{
    const unsigned int maxTickLen = 1000;

    d_data->minLen = qwtMin(minLen, maxTickLen);
    d_data->medLen = qwtMin(medLen, maxTickLen);
    d_data->majLen = qwtMin(majLen, maxTickLen);
}

/*!
    Return the length of the ticks
    \sa QwtScaleDraw::majTickLength()
*/
void QwtScaleDraw::tickLength(unsigned int &minLen,
        unsigned int &medLen, unsigned int &majLen) const
{
    minLen = d_data->minLen;
    medLen = d_data->medLen;
    majLen = d_data->majLen;
}

/*!
    Return the length of the major ticks
    \sa QwtScaleDraw::tickLength()
*/
unsigned int QwtScaleDraw::majTickLength() const
{
    return d_data->majLen;
}

/*!
  \param font Font
  \return the maximum width of a label
*/
int QwtScaleDraw::maxLabelWidth(const QFont &font) const
{
    int maxWidth = 0;

    const QwtTickList &ticks = d_data->scldiv.ticks(QwtScaleDiv::MajorTick);
    for (uint i = 0; i < (uint)ticks.count(); i++)
    {
        const double v = ticks[i];
        if ( d_data->scldiv.contains(v) )
        {
            const int w = labelBoundingRect(font, ticks[i]).width();
            if ( w > maxWidth )
                maxWidth = w;
        }
    }

    return maxWidth;
}

/*!
  \param font Font
  \return the maximum width of a label
*/
int QwtScaleDraw::maxLabelHeight(const QFont &font) const
{
    int maxHeight = 0;

    const QwtTickList &ticks = d_data->scldiv.ticks(QwtScaleDiv::MajorTick);
    for (uint i = 0; i < (uint)ticks.count(); i++)
    {
        const double v = ticks[i];
        if ( d_data->scldiv.contains(v) )
        {
            const int h = labelBoundingRect(font, ticks[i]).height();
            if ( h > maxHeight )
                maxHeight = h;
        }
    }

    return maxHeight;
}

/*!
  Find the bounding rect for the label. The coordinates of
  the rect are relative to margin + ticklength from the backbone
  in direction of the tick.
*/
QRect QwtScaleDraw::labelBoundingRect(
    const QFont &font, double val) const
{
    const QFontMetrics fm(font);

    QString zeroString;
    if ( d_data->fieldwidth > 0 )
        zeroString.fill('0', d_data->fieldwidth);

    const QString lbl = label(val);

    const QString &txt = fm.width(zeroString) > fm.width(lbl) 
        ? zeroString : lbl;
    if ( txt.isEmpty() )
        return QRect(0, 0, 0, 0);

    QRect br;

    QPoint pos;
    int flags;
    double rotation;

    labelPlacement(fm, val, pos, flags, rotation);
    if ( flags )
    {
        // Don't use fm.boundingRect(txt), it cuts off pixels.
        const int w = fm.boundingRect(0, 0, 
            QWIDGETSIZE_MAX, QWIDGETSIZE_MAX, 0, txt).width();
        const int h = -(fm.ascent() - 2);

        QwtMatrix m = labelWorldMatrix(fm, pos, flags, rotation, txt);
        br = QwtMetricsMap::translate(m, QRect(0, 0, w, h));
#if QT_VERSION < 0x040000
        br.moveBy(-pos.x(), -pos.y());
#else
        br.translate(-pos.x(), -pos.y());
#endif
    }

    return br;
}

/*!
  \brief Determine the minimum border distance

  This member function returns the minimum space
  needed to draw the mark labels at the scale's endpoints.

  \param font Font
  \param start start border distance
  \param end end border distance
*/
void QwtScaleDraw::minBorderDist(const QFont &font,
    int &start, int &end ) const
{
    start = 0;
    end = 0;

    const QwtTickList &ticks = d_data->scldiv.ticks(QwtScaleDiv::MajorTick);
    if ( ticks.count() > 0 )
    {
#ifdef __GNUC__
#warning Wrong calculation for scales without bounding labels
#endif

        const QRect labelRectMin = labelBoundingRect(font, ticks[0]);
        const QRect labelRectMax = labelBoundingRect(font, 
            ticks[ticks.count() - 1]);

        switch (d_data->orient)
        {
            case Left:
            case Right:
                end = -labelRectMin.y();
                start = labelRectMax.height() + labelRectMax.y();
                break;
            case Top:
            case Bottom:
                start = -labelRectMin.x();
                end = labelRectMax.width() + labelRectMax.x();
                break;
            case Round:
                start = labelRectMin.width();
                end = labelRectMax.width();
                break;
        }
    }
}

/*!
  Determine the minimum distance between two labels, that is necessairy
  that the texts don't overlap.

  \param font Font
  \return the maximum width of a label
  \warning Not implemented for round scales
*/

int QwtScaleDraw::minLabelDist(const QFont &font) const
{
    if ( d_data->orient == Round ) // no implementation
        return 0;

    const QwtTickList &ticks = d_data->scldiv.ticks(QwtScaleDiv::MajorTick);
    if (ticks.count() == 0)
        return 0;

    const QFontMetrics fm(font);

    const bool vertical = (d_data->orient == Left || d_data->orient == Right);

    QRect bRect1;
    QRect bRect2 = labelBoundingRect(font, ticks[0]);
    if ( vertical )
    {
        bRect2.setRect(-bRect2.bottom(), 0, bRect2.height(), bRect2.width());
    }
    int maxDist = 0;

    for (uint i = 1; i < (uint)ticks.count(); i++ )
    {
        bRect1 = bRect2;
        bRect2 = labelBoundingRect(font, ticks[i]);
        if ( vertical )
        {
            bRect2.setRect(-bRect2.bottom(), 0,
                bRect2.height(), bRect2.width());
        }

        int dist = fm.leading(); // space between the labels
        if ( bRect1.right() > 0 )
            dist += bRect1.right();
        if ( bRect2.left() < 0 )
            dist += -bRect2.left();

        if ( dist > maxDist )
            maxDist = dist;
    }

    double angle = d_data->labelRotation / 180.0 * M_PI;
    if ( vertical )
        angle += M_PI / 2;

    if ( sin(angle) == 0.0 )
        return maxDist;

    const int fmHeight = fm.ascent() - 2; 

    // The distance we need until there is
    // the height of the label font. This height is needed
    // for the neighbour labal.

    int labelDist = (int)(fmHeight / sin(angle) * cos(angle));
    if ( labelDist < 0 )
        labelDist = -labelDist;

    // The cast above floored labelDist. We want to ceil.
    labelDist++; 

    // For text orientations close to the scale orientation 

    if ( labelDist > maxDist )
        labelDist = maxDist;

    // For text orientations close to the opposite of the 
    // scale orientation

    if ( labelDist < fmHeight )
        labelDist = fmHeight;

    return labelDist;
}

/*!
  \param pen pen
  \param font Font
  \return the minimum height required to draw the scale
          including the minimum border distance
*/
int QwtScaleDraw::minHeight( const QPen &pen, const QFont &font ) const
{
    const int pw = qwtMax( 1, pen.width() );  // penwidth can be zero

    int h = 0;
    switch ( d_data->orient )
    {
        case Left:
        case Right:
        {
            const uint minorCount = 
                d_data->scldiv.ticks(QwtScaleDiv::MinorTick).count() +
                d_data->scldiv.ticks(QwtScaleDiv::MediumTick).count();
            const uint majorCount = 
                d_data->scldiv.ticks(QwtScaleDiv::MajorTick).count();

            int bottomDist, topDist;
            minBorderDist(font, bottomDist, topDist);

            h = bottomDist + topDist;
            if ( majorCount >= 2 ) 
                h += minLabelDist(font) * (majorCount - 1);

            int th = 2 * (majorCount + minorCount) * pw;
            if ( th > h )
                h = th;
            break;
        }
        case Round:
            // compute the radial thickness
            h = pw + d_data->vpad + d_data->majLen + maxLabelWidth(font);
            break;
        case Top:
        case Bottom:
            h = pw + d_data->vpad + d_data->majLen + maxLabelHeight(font);
            break;
    }
    return h;
}

/*!
  \param pen pen
  \param font Font
  \return the minimum width required to draw the scale
          including the minimum border distance
*/
int QwtScaleDraw::minWidth( const QPen &pen, const QFont &font ) const
{
    const int pw = qwtMax( 1, pen.width() );  // penwidth can be zero

    int w = 0;

    switch(d_data->orient)
    {
        case Left:
        case Right:
        {
            w = pw + d_data->hpad + d_data->majLen + maxLabelWidth(font);
            break;
        }
        case Round:
        {   
            w = pw + d_data->vpad + d_data->majLen + maxLabelWidth(font);
            break;
        }
        case Top:
        case Bottom:
        {
            const uint minorCount =
                d_data->scldiv.ticks(QwtScaleDiv::MinorTick).count() +
                d_data->scldiv.ticks(QwtScaleDiv::MediumTick).count();
            const uint majorCount =
                d_data->scldiv.ticks(QwtScaleDiv::MajorTick).count();

            int leftDist, rightDist;
            minBorderDist(font, leftDist, rightDist);

            w = leftDist + rightDist + 
                minLabelDist(font) * (majorCount - 1);

            int tw = 2 * (majorCount + minorCount) * pw;
            if ( tw > w )
                w = tw;

            break;
        }
    }
    return w;
}

/*!
  \brief Convert a value into its representing label using the labelFormat. 
  \param value Value
  \return Label string.
  \sa QwtScaleDraw::setLabelFormat()
*/

QString QwtScaleDraw::label(double value) const
{
    // snprintf is C99 and therefore not portable :-(
    // sprintf is safe, when buffer is large enough for
    // a minus sign, 99 digits, a radix symbol,
    // an exponent character, a minus or plus sign, 3 digits,
    // and a '\0'.
    char buffer[128];
    sprintf(buffer, d_data->formatBuffer, value);

    return QString::fromLatin1(buffer);
}

//! Return x origin
int QwtScaleDraw::x() const
{
    return d_data->xorg;
}

//! Return y origin
int QwtScaleDraw::y() const
{
    return d_data->yorg;
}

//! Return length
int QwtScaleDraw::length() const
{
    return d_data->len;
}

//! Return scale orientation 
QwtScaleDraw::Orientation QwtScaleDraw::orientation() const 
{ 
    return d_data->orient; 
}
