/* -*- mode: C++ ; c-file-style: "stroustrup" -*- *****************************
 * Qwt Widget Library
 * Copyright (C) 1997   Josef Wilgen
 * Copyright (C) 2002   Uwe Rathmann
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the Qwt License, Version 1.0
 *****************************************************************************/

// vim: expandtab

#include <qpainter.h>
#include <qevent.h>
#include "qwt_painter.h"
#include "qwt_scale_widget.h"
#include "qwt_math.h"
#include "qwt_paint_buffer.h"
#include "qwt_scale_div.h"
#include "qwt_text.h"

class QwtScaleWidget::PrivateData
{
public:
    PrivateData():
        scaleDraw(NULL),
        title(NULL)
    {
    }

    ~PrivateData()
    {
        delete scaleDraw;
        delete title;
    }

    QwtScaleDraw *scaleDraw;

    int borderDist[2];
    int scaleLength;
    int baseDist;

    int titleOffset;
    int titleDist;
    QwtText *title;
};

/*!
  \brief Create a scale with the position QwtScaleWidget::Left
  \param parent
  \param name passed to QWidget's CTOR

  \warning Workaround of a designer problem
*/
QwtScaleWidget::QwtScaleWidget(QWidget *parent):
    QWidget(parent)
{
    initScale(QwtScaleWidget::Left);
}

/*!
  \brief Constructor
  \param pos Position. Must be one of
     <code>QwtScaleWidget::Left, QwtScaleWidget::Right, QwtScaleWidget::Bottom,
     </code>or \c QwtScaleWidget::Top
  \param parent
  \param name passed to QWidget's CTOR
*/
QwtScaleWidget::QwtScaleWidget(QwtScaleWidget::Position pos, QWidget *parent):
    QWidget(parent)
{
    initScale(pos);
}

//! Destructor
QwtScaleWidget::~QwtScaleWidget()
{
    delete d_data;
}

//! Initialize the scale
void QwtScaleWidget::initScale(QwtScaleWidget::Position pos)
{
    d_data = new PrivateData;

#if QT_VERSION < 0x040000
    setWFlags(Qt::WNoAutoErase);
#endif 

    QwtScaleDraw::Orientation sdo;
    d_data->borderDist[0] = 0;
    d_data->borderDist[1] = 0;
    d_data->baseDist = 4;
    d_data->titleOffset = 0;
    d_data->titleDist = 2;

    switch(pos)
    {
        case QwtScaleWidget::Left:
            sdo = QwtScaleDraw::Left;
            break;
        case QwtScaleWidget::Right:
            sdo = QwtScaleDraw::Right;
            break;
        case QwtScaleWidget::Bottom:
            sdo = QwtScaleDraw::Bottom;
            break;
        case QwtScaleWidget::Top:
            sdo = QwtScaleDraw::Top;
            break;
        default:
            sdo = QwtScaleDraw::Left;
    }

    d_data->scaleDraw = new QwtScaleDraw;
    d_data->scaleDraw->setGeometry(0, 0, 10, sdo);

    const int flags = Qt::AlignHCenter
#if QT_VERSION < 0x040000
        | Qt::WordBreak | Qt::ExpandTabs;
#else
        | Qt::TextExpandTabs | Qt::TextWordWrap;
#endif
    d_data->title = new QwtPlainText( QString::null, font(), flags);

    QSizePolicy policy(QSizePolicy::MinimumExpanding,
        QSizePolicy::Fixed);
    if ( d_data->scaleDraw->orientation() == QwtScaleDraw::Left ||
        d_data->scaleDraw->orientation() == QwtScaleDraw::Right )
    {
        policy.transpose();
    }

    setSizePolicy(policy);
    
#if QT_VERSION >= 0x040000
    setAttribute(Qt::WA_WState_OwnSizePolicy, false);
#else
    clearWState( WState_OwnSizePolicy );
#endif

}

/*!
  \brief Give title new text contents
  \param text New text
  \sa QwtScaleWidget::title
*/
void QwtScaleWidget::setTitle(const QString &text)
{
    if (text == title())
        return;

    QwtText *title = QwtText::makeText(text, d_data->title->flags(),
        d_data->title->font(), d_data->title->color());

    delete d_data->title;
    d_data->title = title;

    layoutScale();
}

/*!
  \brief Assign a position
  \param pos new position
  \sa QwtScaleWidget::position
*/
void QwtScaleWidget::setPosition(Position pos)
{
    QwtScaleDraw::Orientation sdo;
    switch(pos)
    {
        case QwtScaleWidget::Left:
            sdo = QwtScaleDraw::Left;
            break;
        case QwtScaleWidget::Right:
            sdo = QwtScaleDraw::Right;
            break;
        case QwtScaleWidget::Bottom:
            sdo = QwtScaleDraw::Bottom;
            break;
        case QwtScaleWidget::Top:
            sdo = QwtScaleDraw::Top;
            break;
        default:
            sdo = QwtScaleDraw::Left;
    }
    
#if QT_VERSION >= 0x040000
    if ( !testAttribute(Qt::WA_WState_OwnSizePolicy) )
#else
    if ( !testWState( WState_OwnSizePolicy ) )
#endif
    {
        QSizePolicy policy(QSizePolicy::MinimumExpanding,
            QSizePolicy::Fixed);
        if ( d_data->scaleDraw->orientation() == QwtScaleDraw::Left ||
            d_data->scaleDraw->orientation() == QwtScaleDraw::Right )
        {
            policy.transpose();
        }
        setSizePolicy(policy);

#if QT_VERSION >= 0x040000
        setAttribute(Qt::WA_WState_OwnSizePolicy, false);
#else
        clearWState( WState_OwnSizePolicy );
#endif
    }

    if (d_data->scaleDraw)
        d_data->scaleDraw->setGeometry(0,0,10,sdo);
    layoutScale();
}

        
/*! 
    \return position 
    \sa QwtScaleWidget::setPosition
*/
QwtScaleWidget::Position QwtScaleWidget::position() const 
{
    if (!scaleDraw())
        return Left;
    
    QwtScaleDraw::Orientation ori = scaleDraw()->orientation();
    switch(ori)
    {
        case QwtScaleDraw::Left:
            return QwtScaleWidget::Left;
        case QwtScaleDraw::Right:
            return QwtScaleWidget::Right;
        case QwtScaleDraw::Bottom:
            return QwtScaleWidget::Bottom;
        case QwtScaleDraw::Top:
            return QwtScaleWidget::Top;
        default:
            return QwtScaleWidget::Left;
    }
}

/*!
  Specify distances of the scale's endpoints from the
  widget's borders. The actual borders will never be less
  than minimum border distance.
  \param dist1 Left or top Distance
  \param dist2 Right or bottom distance
  \sa QwtScaleWidget::borderDist
*/
void QwtScaleWidget::setBorderDist(int dist1, int dist2)
{
    if ( dist1 != d_data->borderDist[0] || dist2 != d_data->borderDist[1] )
    {
        d_data->borderDist[0] = dist1;
        d_data->borderDist[1] = dist2;
        layoutScale();
    }
}

/*!
  \brief Specify the border distance of the scale's base line.
  \param bd border distance of the baseline
  \sa QwtScaleWidget::baseLineDist
*/
void QwtScaleWidget::setBaselineDist(int bd)
{
    bd = qwtMax( 0, bd );
    if ( bd != d_data->baseDist )
    {
        d_data->baseDist = bd;
        layoutScale();
    }
}

/*!
  \brief Specify the distance between scale and title
  \param td Title distance
  \sa QwtScaleWidget::titleDist
*/
void QwtScaleWidget::setTitleDist(int td)
{
    td = qwtMax( 0, td );
    if ( td != d_data->titleDist )
    {
        d_data->titleDist = td;
        layoutScale();
    }
}


/*!
  \brief Change the title font
  \param f new font
  \sa QwtScaleWidget::title
*/
void QwtScaleWidget::setTitleFont(const QFont &f)
{
    if (f == d_data->title->font())
        return;

    d_data->title->setFont(f);
    layoutScale();
}

/*!
  \brief Change the title color
  \param c new title color
  \sa QwtScaleWidget::titleColor
*/
void QwtScaleWidget::setTitleColor(const QColor &c)
{
    d_data->title->setColor(c);
    update();
}

/*!
  \brief Change the title flags
  \param flags see Qt::AlignmentFlags, Qt::TextFlags. flags are interpreted in
               direction of the label, AlignTop, AlignBottom can't be set
               as the title will always be aligned to the scale.
  \sa QwtScaleWidget::titleFlags, Qt::AlignmentFlags, Qt::TextFlags

*/
void QwtScaleWidget::setTitleFlags(int flags)
{
    flags &= ~(Qt::AlignTop | Qt::AlignBottom);
    d_data->title->setFlags(flags);
    layoutScale(); // Qt::SingleLine, Qt::ExpandTabs might affect layout
}

/*!
  \brief Set the number format for the major scale labels

  Format character and precision have the same meaning as for the
  QString class.
  \param f  format character
  \param prec precision
  \param fieldwidth minimum fieldwidth
  \sa QwtScaleWidget::labelFormat, QString::sprintf in the Qt manual
*/

void QwtScaleWidget::setLabelFormat(char f, int prec, int fieldwidth)
{
    d_data->scaleDraw->setLabelFormat(f, prec, fieldwidth);
    layoutScale();
}

/*!
  \brief Change the flags for the labels.

  See QwtScaleDraw::setLabelFlags().
  \sa QwtScaleDraw::setLabelFlags(), QwtScaleWidget::setLabelRotation()
*/

void QwtScaleWidget::setLabelFlags(int flags)
{
    d_data->scaleDraw->setLabelFlags(flags);
    layoutScale();
}

/*!
  \brief Change the rotation for the labels.
  See QwtScaleDraw::setLabelRotation().
  \sa QwtScaleDraw::setLabelRotation(), QwtScaleWidget::setLabelFlags()
*/
void QwtScaleWidget::setLabelRotation(double rotation)
{
    d_data->scaleDraw->setLabelRotation(rotation);
    layoutScale();
}

/*!
  \brief Set a scale draw
  sd has to be created with new and will be deleted in
  QwtScaleWidget::~QwtScale or the next call of QwtScaleWidget::setScaleDraw.
*/
void QwtScaleWidget::setScaleDraw(QwtScaleDraw *sd)
{
    if ( sd == NULL || sd == d_data->scaleDraw )
        return;

    if ( d_data->scaleDraw )
        *sd = *d_data->scaleDraw;

    delete d_data->scaleDraw;
    d_data->scaleDraw = sd;

    layoutScale();
}

/*!
  Format character and precision have the same meaning as for the
  QString class.
  \retval f format character
  \retval prec precision
  \retval fieldwidth minimum fieldwidth
  \return the number format for the major scale labels
  \sa QwtScaleWidget::setLabelFormat, QString::sprintf in the Qt manual
*/

void QwtScaleWidget::labelFormat(char &f, int &prec, int &fieldwidth) const
{
    d_data->scaleDraw->labelFormat(f, prec, fieldwidth);
}

/*! 
    scaleDraw of this scale
    \sa QwtScaleDraw::setScaleDraw
*/
const QwtScaleDraw *QwtScaleWidget::scaleDraw() const 
{ 
    return d_data->scaleDraw; 
}

/*! 
    \return alignment
    \sa QwtScaleWidget::setTitleFlags, Qt::AlignmentFlags, Qt::TextFlags
*/
int QwtScaleWidget::titleFlags() const 
{ 
    return d_data->title->flags();
}

/*! 
    \return title font
    \sa QwtScaleWidget::setTitleFont
*/
const QFont QwtScaleWidget::titleFont() const 
{
    return d_data->title->font();
}

/*! 
    \return title color
    \sa QwtScaleWidget::setTitleColor
*/
const QColor QwtScaleWidget::titleColor() const 
{
    return d_data->title->color();
}

/*! 
    \return title 
    \sa QwtScaleWidget::setTitle
*/
const QString QwtScaleWidget::title() const 
{
    return d_data->title->text();
}

/*! 
    \return start border distance 
    \sa QwtScaleWidget::setBorderDist
*/
int QwtScaleWidget::startBorderDist() const 
{ 
    return d_data->borderDist[0]; 
}  

/*! 
    \return end border distance 
    \sa QwtScaleWidget::setBorderDist
*/
int QwtScaleWidget::endBorderDist() const 
{ 
    return d_data->borderDist[1]; 
}

/*! 
    \return base line distance 
    \sa QwtScaleWidget::setBaselineDist
*/
int QwtScaleWidget::baseLineDist() const 
{ 
    return d_data->baseDist; 
}

/*! 
    \return distance between scale and title
    \sa QwtScaleWidget::setBaselineDist
*/
int QwtScaleWidget::titleDist() const 
{ 
    return d_data->titleDist; 
}

/*!
  \brief paintEvent
*/
void QwtScaleWidget::paintEvent(QPaintEvent *e)
{
    const QRect &ur = e->rect();
    if ( ur.isValid() )
    {
#if QT_VERSION < 0x040000
        QwtPaintBuffer paintBuffer(this, ur);
        draw(paintBuffer.painter());
#else
        QPainter painter(this);
        draw(&painter);
#endif
    }
}

/*!
  \brief draw the scale
*/
void QwtScaleWidget::draw(QPainter *p) const
{
#if QT_VERSION < 0x040000
    d_data->scaleDraw->draw(p, colorGroup());
#else
    d_data->scaleDraw->draw(p, palette());
#endif

    QRect r = rect();

    switch(d_data->scaleDraw->orientation())
    {
        case QwtScaleDraw::Bottom:
            r.setTop( r.top() + d_data->titleOffset );
            break;

        case QwtScaleDraw::Left:
            r.setRight( r.right() - d_data->titleOffset );
            break;

        case QwtScaleDraw::Right:
            r.setLeft( r.left() + d_data->titleOffset );
            break;

        case QwtScaleDraw::Top:
        default:
            r.setBottom( r.bottom() - d_data->titleOffset );
            break;
    }

    drawTitle(p, d_data->scaleDraw->orientation(), r);
}

/*!
  \brief resizeEvent
*/
void QwtScaleWidget::resizeEvent(QResizeEvent *)
{
    layoutScale(false);
}

//! Recalculate the scale's geometry and layout based on
//  the current rect and fonts.
//  \param update_geometry   notify the layout system and call update
//         to redraw the scale

void QwtScaleWidget::layoutScale( bool update_geometry )
{
    QRect r = this->rect();

    // Borders cannot be less than the minBorderDist

    int bd0, bd1;
    minBorderDist(bd0, bd1);
    if ( d_data->borderDist[0] > bd0 )
        bd0 = d_data->borderDist[0];
    if ( d_data->borderDist[1] > bd1 )
        bd1 = d_data->borderDist[1];

    switch(d_data->scaleDraw->orientation())
    {
    case QwtScaleDraw::Bottom:
        d_data->scaleDraw->setGeometry(r.x() + bd0, r.y() + d_data->baseDist,
            r.width() - bd0 - bd1, QwtScaleDraw::Bottom);

        d_data->titleOffset = d_data->titleDist + d_data->baseDist +
            d_data->scaleDraw->maxHeight(QPen(), font());
        break;

    case QwtScaleDraw::Top:
        d_data->scaleDraw->setGeometry(r.x() + bd0, r.bottom() - d_data->baseDist,
               r.width() - bd0 - bd1, QwtScaleDraw::Top);
        d_data->titleOffset =  d_data->titleDist + d_data->baseDist +
                           d_data->scaleDraw->maxHeight(QPen(), font());
        break;

    case QwtScaleDraw::Left:
        d_data->scaleDraw->setGeometry(r.right() - d_data->baseDist, r.y() + bd0,
                            r.height() - bd0 - bd1, QwtScaleDraw::Left);
        d_data->titleOffset = d_data->baseDist + d_data->titleDist +
                        d_data->scaleDraw->maxWidth(QPen(), font());
        break;

    case QwtScaleDraw::Right:
        d_data->scaleDraw->setGeometry(r.x() + d_data->baseDist, r.y() + bd0,
                            r.height() - bd0 - bd1, QwtScaleDraw::Right);
        d_data->titleOffset = d_data->baseDist + d_data->titleDist +
                          d_data->scaleDraw->maxWidth(QPen(), font());
        break;
    default:
        break;
    }
    if ( update_geometry )
    {
      updateGeometry();
      update();
    }
}

/*!
  Rotate and paint a title according to its position into a given rectangle.
  \param painter Painter
  \param o Orientation
  \param rect Bounding rectangle
*/

void QwtScaleWidget::drawTitle(QPainter *painter,
    QwtScaleDraw::Orientation o, const QRect &rect) const
{
    QRect r;
    double angle;
    int flags = d_data->title->flags() & ~(Qt::AlignTop | Qt::AlignBottom);

    switch(o)
    {
#ifndef QT_NO_TRANSFORMATIONS
        case QwtScaleDraw::Left:
            flags |= Qt::AlignTop;
            angle = -90.0;
            r.setRect(rect.left(), rect.bottom(), rect.height(), rect.width());
            break;
        case QwtScaleDraw::Right:
            flags |= Qt::AlignTop;
            angle = 90.0;
            r.setRect(rect.right(), rect.top(), rect.height(), rect.width());
            break;
#else
        case QwtScaleDraw::Left:
        case QwtScaleDraw::Right:
            // swallow vertical titles
            return;
#endif
        case QwtScaleDraw::Top:
            flags |= Qt::AlignTop;
            angle = 0.0;
            r = rect;
            break;
        case QwtScaleDraw::Bottom:
        default:
            flags |= Qt::AlignBottom;
            angle = 0.0;
            r = rect;
            break;
    }

    d_data->title->setFlags(flags);

    painter->save();
    painter->translate(r.x(), r.y());
#ifndef QT_NO_TRANSFORMATIONS
    if (angle != 0.0)
        painter->rotate(angle);
#endif
    d_data->title->draw(painter, QRect(0, 0, r.width(), r.height()));

    // Users do not expect to see AlignTop or AlignBottom
    d_data->title->setFlags(flags & ~(Qt::AlignTop | Qt::AlignBottom));
    painter->restore();
}

/*!
  \brief Notify a change of the scale

  This virtual function can be overloaded by derived
  classes. The default implementation updates the geometry
  and repaints the widget.
*/

void QwtScaleWidget::scaleChange()
{
    layoutScale();
}

/*!
  \return a size hint
*/
QSize QwtScaleWidget::sizeHint() const
{
    return minimumSizeHint();
}

/*!
  \return a minimum size hint
*/
QSize QwtScaleWidget::minimumSizeHint() const
{
    Qt::Orientation o = Qt::Horizontal;
    if ( d_data->scaleDraw->orientation() == QwtScaleDraw::Left ||
        d_data->scaleDraw->orientation() == QwtScaleDraw::Right )
    {
        o = Qt::Vertical;
    }

    // Border Distance cannot be less than the scale minBorderDist
    // Note, the minBorderDist is already included in minHeight/minWidth
    int length = 0;
    int mbd1, mbd2;
    minBorderDist(mbd1, mbd2);
    length += qwtMax( 0, d_data->borderDist[0] - mbd1 );
    length += qwtMax( 0, d_data->borderDist[1] - mbd2 );

    if ( o == Qt::Vertical )
        length += d_data->scaleDraw->minHeight(QPen(), font());
    else
        length += d_data->scaleDraw->minWidth(QPen(), font());

    int dim = dimForLength(length, font());
    if ( length < dim )
    {
        // compensate for long titles
        length = dim;
        dim = dimForLength(length, font());
    }

    QSize size(length + 2, dim);
    if ( o == Qt::Vertical )
        size.transpose();

    return size;
}

/*!
  \brief Find the height of the title for a given width.
  \param width Width
  \return height Height
 */

int QwtScaleWidget::titleHeightForWidth(int width) const
{
    return d_data->title->heightForWidth(width);
}

/*!
  \brief Find the minimum dimension for a given length.
         dim is the height, length the width seen in
         direction of the title.
  \param length width for horizontal, height for vertical scales
  \param scaleFont Font of the scale
  \return height for horizontal, width for vertical scales
*/

int QwtScaleWidget::dimForLength(int length, const QFont &scaleFont) const
{
    int dim = d_data->baseDist;
    if ( d_data->scaleDraw->orientation() == QwtScaleDraw::Left ||
        d_data->scaleDraw->orientation() == QwtScaleDraw::Right )
    {
        dim += d_data->scaleDraw->minWidth(QPen(), scaleFont);
#ifdef QT_NO_TRANSFORMATIONS
        return dim;
#endif
    }
    else
        dim += d_data->scaleDraw->minHeight(QPen(), scaleFont);

    if ( !d_data->title->text().isEmpty() )
    {
        dim += titleHeightForWidth(length) + d_data->titleDist;
    }
    return dim;
}

/*!
  \brief Determine the minimum Border distances.

  This member function returns the minimum distance
  of the scale's endpoints from the widget borders which
  is required for the mark labels to fit into the widget.
  \warning
  <ul> <li>The minimum border distance depends on the font.</ul>
  \sa QwtScaleWidget::setBorderDist()
*/
void QwtScaleWidget::minBorderDist(int &start, int &end) const
{
    d_data->scaleDraw->minBorderDist(font(), start, end);
}

/*!
  \brief Notify a change of the font

  This virtual function may be overloaded by derived widgets.
  The default implementation resizes the scale and repaints
  the widget.
  \param oldFont Previous font
*/
void QwtScaleWidget::fontChange(const QFont &oldFont)
{
    QWidget::fontChange( oldFont );
    layoutScale();
}

/*!
  \brief Assign a scale division

  The scale division determines where to set the tick
  marks.
  \param sd Scale Division
  \sa For more information about scale divisions, see QwtScaleDiv.
*/
void QwtScaleWidget::setScaleDiv(const QwtScaleTransformation& transformation,
    const QwtScaleDiv &sd)
{
    if (d_data->scaleDraw->scaleDiv() != sd)
    {
        d_data->scaleDraw->setTransformation(transformation);
        d_data->scaleDraw->setScaleDiv(sd);
        layoutScale();
    }
}
