/* -*- mode: C++ ; c-file-style: "stroustrup" -*- *****************************
 * Qwt Widget Library
 * Copyright (C) 1997   Josef Wilgen
 * Copyright (C) 2003   Uwe Rathmann
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the Qwt License, Version 1.0
 *****************************************************************************/

// vim: expandtab

#include <qpainter.h>
#include <qpalette.h>
#if QT_VERSION < 0x040000
#include <qsimplerichtext.h>
#include <qstylesheet.h>
#else
#include <qtextdocument.h>
#endif
#include "qwt_painter.h"
#include "qwt_text.h"

Qt::TextFormat QwtText::d_defaultFormat = Qt::AutoText;

class QwtText::PrivateData
{
public:
    PrivateData(const QString &txt, const QFont &fnt, int flgs, 
            const QColor &c, const QPen &pen, const QBrush &brush):
        flags(flgs),
        text(txt),
        font(fnt),
        color(c),
        fontMetrics(fnt),
        rectPen(pen),
        rectBrush(brush)
    {
    }

    int flags;
    QString text;
    QFont font;
    QColor color;
    QFontMetrics fontMetrics;
    QPen rectPen;
    QBrush rectBrush;
};

class QwtRichText::PrivateData
{
public:
    PrivateData():
        doc(NULL)
    {
    }
    ~PrivateData()
    {
        delete doc;
    }
#if QT_VERSION < 0x040000
    QSimpleRichText *doc;
#else
    QTextDocument *doc;
#endif
};

#if QT_VERSION < 0x040000
static QSimpleRichText *richText(const QString &text, const QFont &font)
{
    return new QSimpleRichText(text, font);
}
#else
static QTextDocument *richText(const QString &text, const QFont &font)
{
    QTextDocument *doc = new QTextDocument();
    doc->setUndoRedoEnabled(false);
    doc->setDefaultFont(font);
    doc->setHtml(text);
    
    return doc;
}
#endif

/*!
  \brief Constructor

  \param text Text
  \param font Font
  \param flags Or'd Qt::Alignment or Qt::TextFlags flags
  \param color Color
  \param pen Pen
  \param brush Brush
*/

QwtText::QwtText(const QString &text, const QFont &font, 
   int flags, const QColor &color, const QPen &pen, const QBrush &brush)
{
    d_data = new PrivateData(text, font, flags, color, pen, brush);
}

//! Destructor
QwtText::~QwtText() 
{
    delete d_data;
}

/*!
  Factory function for QwtText objects. Creates a text according
  to QwtText::defaultFormat()

  \param text Contents of the text
  \param flags Or'd Qt::Alignment or Qt::TextFlags flags
  \param font Font of the text
  \param color Color of the text
  \param pen Pen to draw the enclosing rectangle
  \param brush Brush to draw the enclosing rectangle 
  \return Pointer to a new QwtPlainText or QwtRichText object
  \sa QwtText::defaultFormat, QwtText::setDefaultFormat
*/
QwtText *QwtText::makeText(const QString &text, 
    int flags, const QFont &font, const QColor &color,
    const QPen &pen, const QBrush &brush)
{
    return makeText(text, defaultFormat(), flags, font, color, pen, brush);
}

/*!
  Factory function for QwtText objects.

  \param text Contents of the text
  \param format Format of the text
  \param flags Or'd Qt::Alignment or Qt::TextFlags flags
  \param font Font of the text
  \param color Color of the text
  \param pen Pen to draw the enclosing rectangle
  \param brush Brush to draw the enclosing rectangle 
  \return Pointer to a new QwtPlainText or QwtRichText object
*/

QwtText *QwtText::makeText(const QString &text, Qt::TextFormat format,
    int flags, const QFont &font, const QColor &color,
    const QPen &pen, const QBrush &brush)
{
#ifndef QT_NO_RICHTEXT
    bool richText = false;
    if (format == Qt::RichText)
        richText = true;
    else if ( format == Qt::AutoText )
    {
#if QT_VERSION < 0x040000
        if ( QStyleSheet::mightBeRichText(text) )
#else
        if ( Qt::mightBeRichText(text) )
#endif
            richText = true;
    }
    
    if ( richText )
        return new QwtRichText(text, font, flags, color, pen, brush);

#endif

    return new QwtPlainText(text, font, flags, color, pen, brush);
}

//!  Set the default format for the QwtText factory
void QwtText::setDefaultFormat(Qt::TextFormat format)
{
    d_defaultFormat = format;
}

//! Return default format for the QwtText factory
Qt::TextFormat QwtText::defaultFormat()
{
    return d_defaultFormat;
}

//! Set the text.
void QwtText::setText(const QString &text) 
{ 
    d_data->text = text; 
}

//! Return the text.
QString QwtText::text() const 
{ 
    return d_data->text; 
}

//! Set the font.
void QwtText::setFont(const QFont &font) 
{
    d_data->font = font; 
    d_data->fontMetrics = QFontMetrics(font);
}

//! Return the font.
QFont QwtText::font() const 
{ 
    return d_data->font; 
}

//! Return the font metrics.
QFontMetrics QwtText::fontMetrics() const 
{ 
    return d_data->fontMetrics; 
}

/*!
  Set the flags.
  \param flags Or'd Qt::Alignment or Qt::TextFlags flags
*/
void QwtText::setFlags(int flags) 
{ 
    d_data->flags = flags; 
}

//! Get the flags.
int QwtText::flags() const 
{ 
    return d_data->flags; 
}

//! Set the color.
void QwtText::setColor(const QColor &color) 
{ 
    d_data->color = color; 
}

//! Return the color.
QColor QwtText::color() const 
{ 
    return d_data->color; 
}

//! Set the rectangle pen.
void QwtText::setRectPen(const QPen &pen) 
{ 
    d_data->rectPen = pen; 
}

//! Return the rectangle pen.
QPen QwtText::rectPen() const 
{ 
    return d_data->rectPen; 
}

//! Set the rectangle brush.
void QwtText::setRectBrush(const QBrush &brush) 
{ 
    d_data->rectBrush = brush; 
}

//! Return the rectangle brush.
QBrush QwtText::rectBrush() const 
{ 
    return d_data->rectBrush; 
}

/*!
  \brief Constructor

  \param text Contents of the text
  \param font Font of the text
  \param flags Or'd values of Qt::AlignmentFlags or Qt::TextFlags 
  \param color Color of the text
  \param pen Pen to draw the enclosing rectangle
  \param brush Brush to draw the enclosing rectangle 
*/

QwtPlainText::QwtPlainText(const QString &text, const QFont &font,
        int flags, const QColor &color, const QPen &pen, const QBrush &brush):
    QwtText(text, font, flags, color, pen, brush) 
{
}

/*!
  \brief Constructor
  \param text Text
  \param font Font
  \param color Color
*/

QwtPlainText::QwtPlainText(const QString &text, 
        const QFont &font, const QColor &color):
    QwtText(text, font, 
#if QT_VERSION < 0x040000
        Qt::AlignCenter | Qt::WordBreak | Qt::ExpandTabs, 
#else
        Qt::AlignCenter | Qt::TextWordWrap | Qt::TextExpandTabs,
#endif
        color) 
{
}

//! Create a clone
QwtText *QwtPlainText::clone() const
{
    return new QwtPlainText(
        text(), font(), flags(), color(), rectPen(), rectBrush());
}

/*!
  \brief Find the height for a given width.
  \param width Width
  \return height Height 
*/
int QwtPlainText::heightForWidth(int width) const
{
    const QwtLayoutMetrics metrics(QwtPainter::metricsMap());
    return metrics.heightForWidth(text(), width, flags(), fontMetrics());
}

/*!
  \brief Draw the text in a clipping rectangle.
  \param painter Painter
  \param rect Clipping rectangle
*/
void QwtPlainText::draw(QPainter *painter, const QRect &rect) const
{
    painter->save();
    painter->setPen(rectPen());
    painter->setBrush(rectBrush());
    QwtPainter::drawRect(painter, rect);
    painter->restore();

    painter->save();
    painter->setFont(font());
    painter->setPen(color());
    QwtPainter::drawText(painter, rect, flags(), text());
    painter->restore();
}

QRect QwtPlainText::boundingRect(QPainter *painter) const
{
    const QwtLayoutMetrics metrics(QwtPainter::metricsMap());
    
    if (painter)
    {
        painter->save();
        painter->setFont(font());
        const QRect rect = metrics.boundingRect(text(), 
            flags(), painter);
        painter->restore();
        return rect;
    }

    return metrics.boundingRect(text(), flags(), fontMetrics());
}

#ifndef QT_NO_RICHTEXT

/*!
  \brief Constructor

  \param text Contents of the text
  \param font Font of the text
  \param flags Or'd values of Qt::AlignmentFlags, Qt::TextFlags 
  \param color Color of the text
  \param pen Pen to draw the enclosing rectangle
  \param brush Brush to draw the enclosing rectangle 
*/

QwtRichText::QwtRichText(const QString &text, const QFont &font,
        int flags, const QColor &color, const QPen &pen, const QBrush &brush):
    QwtText(text, font, flags, color, pen, brush)
{ 
    d_data = new PrivateData;
    d_data->doc = richText(text, font);
    setText(text); 
}

/*!
  \brief Constructor
  \param text Text
  \param font Font
  \param color Color
*/

QwtRichText::QwtRichText(const QString &text, const QFont &font,
        const QColor &color):
    QwtText(text, font,
#if QT_VERSION < 0x040000
        Qt::AlignCenter | Qt::WordBreak | Qt::ExpandTabs,
#else
        Qt::AlignCenter | Qt::TextWordWrap | Qt::TextExpandTabs,
#endif
        color)
{ 
    d_data = new PrivateData;
    d_data->doc = richText(text, font);
    setText(text); 
}

//! Destructor
QwtRichText::~QwtRichText() 
{ 
    delete d_data; 
}

//! Create a clone
QwtText *QwtRichText::clone() const
{
    return new QwtRichText(
        text(), font(), flags(), color(), rectPen(), rectBrush());
}

//! Set the text
void QwtRichText::setText(const QString &text)
{
    QwtText::setText(text);

    delete d_data->doc;
    d_data->doc = richText(taggedText(text, flags()), font());
}

//! Set the font
void QwtRichText::setFont(const QFont &font)
{
    d_data->doc->setDefaultFont(font);
    QwtText::setFont(font);
}

//! Set the flags
void QwtRichText::setFlags(int flags)
{
    QwtText::setFlags(flags);

    delete d_data->doc;
    d_data->doc = richText(taggedText(text(), flags), font());
}

/*!
  \brief Find the height for a given width
  \param width Width
  \return height Height 
*/
int QwtRichText::heightForWidth(int width) const
{
    const QwtLayoutMetrics metrics(QwtPainter::metricsMap());
    return metrics.heightForWidth(*d_data->doc, width);
}

/*!
  \brief Draw the text in a clipping rectangle
  \param painter Painter
  \param rect Clipping rectangle
*/
void QwtRichText::draw(QPainter *painter, const QRect &rect) const
{
    painter->save();
    painter->setPen(rectPen());
    painter->setBrush(rectBrush());
    QwtPainter::drawRect(painter, rect);
    painter->restore();

    painter->save();

    painter->setPen(color());
    painter->setFont(font());

    QwtPainter::drawSimpleRichText(painter, rect, flags(), *d_data->doc);

    painter->restore();
}

#if QT_VERSION < 0x040000
QRect QwtRichText::boundingRect(QPainter *painter) const
{
    const QwtLayoutMetrics metrics(QwtPainter::metricsMap());
    return metrics.boundingRect(*d_data->doc, flags(), painter);
}
#else
QRect QwtRichText::boundingRect(QPainter *) const
{
    const QwtLayoutMetrics metrics(QwtPainter::metricsMap());
    return metrics.boundingRect(text(), *d_data->doc, flags());
}
#endif

//! Wrap text into <div align=...> </div> tags according align
QString QwtRichText::taggedText(const QString &text, int flags) const
{
    QString rich = text;

    // By default QSimpleRichText is Qt::AlignLeft
    if (flags & Qt::AlignJustify) 
    {
        rich.prepend(QString::fromLatin1("<div align=\"justify\">"));
        rich.append(QString::fromLatin1("</div>"));
    } 
    else if (flags & Qt::AlignRight)
    {
        rich.prepend(QString::fromLatin1("<div align=\"right\">"));
        rich.append(QString::fromLatin1("</div>"));
    }
    else if (flags & Qt::AlignHCenter) 
    {
        rich.prepend(QString::fromLatin1("<div align=\"center\">"));
        rich.append(QString::fromLatin1("</div>"));
    }

    return rich;
}

#endif
