#include <iostream>
#include <sstream>
#include <map>
#include <string>
#include <algorithm>

#include <comedilib.h>
#include <errno.h>
#include <string.h>

#include "ProbeComedi.h"

void ProbeDevices(ComediDeviceMap & devices,
                  ComediHandleMap & handles,
                  unsigned start, unsigned end, bool verbose)
{
  if (verbose)
    std::cerr << "Probing all comedi devices from /dev/comedi" << start << " -> /dev/comedi" << end << std::endl;

  devices.clear();
  CloseProbedHandles(handles);
  handles.clear();
      
  for (unsigned minor = start; minor <= end; ++minor) 
  {
    std::stringstream ss;
    ss << "/dev/comedi" << minor;
    const std::string devfile( ss.str() );
    comedi_t *it = ::comedi_open(devfile.c_str());
    if (!it) {
      if (verbose && (!devices.size() || ::comedi_errno() != ENODEV))
        // only display error message when we haven't a comedi device yet
        // or if it's something other than ENODEV
        std::cerr << devfile << ": " 
                << ::comedi_strerror(::comedi_errno()) << std::endl;
      continue;
    }

    ComediDevice & dev = devices[minor]; // put it in our map, creates a ComediDevice

    // now setup our ComediDevice
    dev.minor = minor;
    ::strncpy(dev.driverName, ::comedi_get_driver_name(it), CD_STRMAX);
    dev.driverName[CD_STRMAX-1] = 0;
    ::strncpy(dev.boardName, ::comedi_get_board_name(it), CD_STRMAX);
    dev.boardName[CD_STRMAX-1] = 0;
    dev.versionCode = ::comedi_get_version_code(it);
    dev.nSubdevs = ::comedi_get_n_subdevices(it);
    
    if (!dev.nSubdevs) {
      std::cerr << "/dev/comedi" << minor << " has 0 subdevices? Ignoring..";
      devices.erase(minor);
      continue;
    }

    handles[minor] = it; // save the handle in the out param

    // now setup each subdevice
    for (unsigned sdevNo = 0; sdevNo < dev.nSubdevs && sdevNo < MAX_SUBDEVS; ++sdevNo) {
      ComediSubDevice & sdev = dev.subdev[sdevNo];
      sdev.minor = minor;
      sdev.id = sdevNo;
      sdev.type = ::comedi_get_subdevice_type(it, sdevNo);
      sdev.nChans = ::comedi_get_n_channels(it, sdevNo);
      
      // HACK for pcmad12/16, which lacks proper range info
      if ( !::strcmp(dev.boardName, "pcmad12")
           || !::strcmp(dev.boardName, "pcmad16") ) {

        sdev.nRanges = 1;
        sdev.range[0].min = 0;
        sdev.range[0].max = 5.0;
        sdev.maxdata = ::comedi_get_maxdata(it, sdevNo, 0);
        
      }  else {

        sdev.maxdata = ::comedi_get_maxdata(it, sdevNo, 0);
        sdev.nRanges = ::comedi_get_n_ranges(it, sdevNo, 0);
        for (unsigned range = 0; range < sdev.nRanges && range < MAX_RANGES; ++range) {
          comedi_range *r = comedi_get_range(it, sdevNo, 0, range);
          if (!r) { sdev.nRanges = range; break; }
          CR_From_comedi_range(sdev.range[range], *r, range);        
        }
      }
    }
      //comedi_close(it); don't close!! we keep the handle 
  }
}

static void CloseComediHandle(const std::pair<unsigned, void *> & p)
{ ::comedi_close(static_cast<comedi_t *>(p.second)); }

extern void CloseProbedHandles(const ComediHandleMap & handles)
{
  // clear any passed-in handles by closing them!
  for_each(handles.begin(), handles.end(), CloseComediHandle);

}

