#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <netinet/in.h>
#include <netinet/ip.h> /* superset of previous */
#include <netinet/tcp.h> 
#include <netdb.h>
extern int h_errno;
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <pthread.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>

#define _COMEDILIB_DEPRECATED /* needed because newer versions of COMEDI deprecated
                                 comedi_dio_bitfield! */
#include <comedilib.h>

#include <iostream>
#include <string>
#include <sstream>
#include <algorithm>
#include <vector>

#include "ProbeComedi.h"
#include "ComediDevice.h"
#include "CID.h"
#include "Protocol.h"

static ComediDeviceMap devices;
static ComediHandleMap handles;

#define DEFAULT_LISTEN "0.0.0.0" // listen on all interfaces by default
#define DEFAULT_PORT "3335"
#define CONN_TIMEOUT (60*60*1000) // 1 hour timeout for conns?
#define MAX_LINE_LEN 2048

struct ThreadArg;

// note not all the maps have a valid ThreadArg, only the acq. threads do
typedef std::map<pthread_t, ThreadArg *> ThreadMap;
typedef std::map<unsigned long long, ThreadArg *> DataThreadKeys; // acq. thread threadmap is put here until the client connects..
typedef std::map<CID, ComediSubDevice::DIOConfig> DioConfigs; // keep track of dio configurations for each channel

static pthread_mutex_t mut = PTHREAD_MUTEX_INITIALIZER;

static ThreadMap threads; // all the threads
static DataThreadKeys dataThreadKeys; // keys for data threads..
static DioConfigs dioConfigs;
static bool debug = false;

struct Log
{
  static pthread_mutex_t mut;

  Log() : output_stream(&std::cout) { pthread_mutex_lock(&mut); }
  template <typename T> Log & operator<<(const T & t) { if (output_stream) { std::ostringstream os; os << data << t;  data = os.str();  }  return *this; }
  ~Log() { if (output_stream) (*output_stream) << data; pthread_mutex_unlock(&mut); }
protected:
  std::string data;
  std::ostream *output_stream;
};

pthread_mutex_t Log::mut = PTHREAD_MUTEX_INITIALIZER;

struct Debug : public Log
{
  Debug() { if (!debug) output_stream = 0; }
};

struct Error : public Log
{
  Error() { output_stream = &std::cerr; }
};

static unsigned short parsePort(const char *p)
{
  unsigned short port = 0;
  std::stringstream ss(p);
  ss >> port;
  return port;
}

static u_int32_t parseAddress(const char *a)
{
  struct hostent *he = gethostbyname(a);
  if (!he) return 0;
  return *(u_int32_t *)he->h_addr;
}

static volatile bool stopItAll = false;


void sighandler(int sig)
{
    (void) sig;
    stopItAll = true;
}


static int BindSocket(int argc, char *argv[])
{
   unsigned short port = 0;
   const char *hostArg = DEFAULT_LISTEN, 
              *portArg = DEFAULT_PORT; 
   
   

   int s = ::socket(PF_INET, SOCK_STREAM, 0);

   if (s < 0) {
     perror("socket");
     exit(1);
   }

   int parm = 1;
   ::setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &parm, sizeof(parm));

   if (argc == 3) {
     hostArg = argv[1];
     portArg = argv[2];
   } else if (argc == 2) {
     portArg = argv[1];
   }
   port = parsePort(portArg);
   
   if (!port) {
       Error() << "Unacceptable port argument \"" << portArg << "\", exiting.\n";
       exit(1);
   }

   struct sockaddr_in sin;
   sin.sin_family = AF_INET;
   sin.sin_port = htons(port);
   sin.sin_addr.s_addr = parseAddress(hostArg);
   if (!sin.sin_addr.s_addr) hostArg = DEFAULT_LISTEN;

   int ret = ::bind(s, reinterpret_cast<struct sockaddr *>(&sin), sizeof(sin));
   if (ret) {
     perror("bind");
     exit(1);
   }

   Log() << "Bound socket to " << hostArg << ":" << port << "\n";  
   
   return s;
}

static int serverSocket = -1;

static void ReleaseSocket(void)
{
    ::close(serverSocket);
}


struct ThreadArg
{
  ThreadArg() 
    : fd(-1), sdev(0), chan(0),  range(0), aref(0), hz(1)  { }

  ~ThreadArg() { }

  int fd; // if negative, it's invalid
  ComediSubDevice * sdev;
  unsigned chan; // for DIO this is a bitmask of channels we want.. 
  unsigned range;
  unsigned aref;
  unsigned hz;
};

static void timespec_add_ns(struct timespec & ts, unsigned long ns)
{
    ts.tv_nsec += ns % 1000000000;
    if (ts.tv_nsec >= 1000000000) {
      ts.tv_nsec -= 1000000000;
      ts.tv_sec += ns /  1000000000 + 1;
    } else 
      ts.tv_sec += ns /  1000000000;          
}

static bool TransmitBuf(int fd, void *buf, size_t num)
{
  ssize_t ret;

  while (num) {
    ret = send(fd, buf, num, MSG_NOSIGNAL);

    if (ret < 0 && errno != EAGAIN) return false;
    else if (ret > 0) {
      num -= ret;
      buf = static_cast<char *>(buf) + ret;
    } else {
      Error() << "Unexpected condition in TransmitBuf() or connection lost!\n";
      pthread_exit(0);
    }
  }
  return true;
}

static void *AcquisitionThread(void *arg)
{
    ThreadArg *ta = reinterpret_cast<ThreadArg *>(arg);
    struct timespec nextWakeup;
    unsigned long taskPeriod = 1000000000/ta->hz; 
    lsampl_t sample;
    comedi_t *handle = (comedi_t *)handles[ta->sdev->minor];    
    unsigned sdev = ta->sdev->id, chan = ta->chan, range = ta->range, aref = ta->aref, type = ta->sdev->type;
    int fd = ta->fd;

    clock_gettime(CLOCK_REALTIME, &nextWakeup);
    
    while (1) {
            
      pthread_testcancel();      
      
      if(type == COMEDI_SUBD_DI || type == COMEDI_SUBD_DIO 
         || type == COMEDI_SUBD_DO) {
        // DI/DO/DIO
        comedi_dio_bitfield(handle, sdev, 0, &sample);
        sample &= chan; // mask out the not-wanted channels
      } else { 
        // generic and/or AI
        // not using commands, just do synchronous reads        
        comedi_data_read(handle, sdev, chan, range, aref, &sample);
      }
      sampleToLE(&sample); // convert to protocol-order (little endian)
      if ( !TransmitBuf(fd, &sample, sizeof(lsampl_t)) ) {
        pthread_exit(0); // xmit error.. EOF?
      }
      
      timespec_add_ns(nextWakeup, taskPeriod);
      clock_nanosleep(CLOCK_REALTIME, TIMER_ABSTIME, &nextWakeup, NULL);
    }
    
    return 0;
}

static void connThreadCleanupHandler(void *arg)
{
    int s = (int)arg;
    
    ::close(s);
    pthread_mutex_lock(&mut);
    ThreadMap::iterator it = threads.find(pthread_self());
    if (it != threads.end()) {
      delete it->second; // if it's a valid pointer
      pthread_detach(pthread_self()); // reap ourselves
      threads.erase(pthread_self());
    }
    pthread_mutex_unlock(&mut);
    Debug() << "-Thread: " << reinterpret_cast<void *>(pthread_self()) << " exiting..\n";
}

static void connThreadLineProc(int sock, std::string & line);

static void *connThread(void *arg)
{
    int s = (int)arg, tmp = 1;

    pthread_cleanup_push(connThreadCleanupHandler, (void *)s);

    Debug() << "+Thread: " << reinterpret_cast<void *>(pthread_self()) << " entered\n";

    // turn off nagle, just to be peppy
    tmp = ::setsockopt(s, IPPROTO_TCP, TCP_NODELAY, &tmp, sizeof(tmp));
    if (tmp) {
      ::perror("setsockopt");
      return NULL;
    }
    
    char line[MAX_LINE_LEN];
    char * newlinePos = 0;
    line[0] = 0;

    while (1) {
      struct pollfd pfd = { fd: s, events: POLLIN|POLLPRI, revents: 0 };

      pthread_testcancel();

      tmp = ::poll(&pfd, 1, CONN_TIMEOUT);

      if (tmp < 0) {
        if (errno == EINTR) continue; // try again?
        ::perror("poll");
        return NULL;
      } else if (tmp == 0) {
        // timeout?
        pthread_exit(0);
      } else {
        // regular read?
        char *newpos = line;
        unsigned sz = 0;
        while (!(newlinePos = strchr(newpos, '\n')))  {
          tmp = ::recv(s, line + sz, MAX_LINE_LEN - sz, MSG_DONTWAIT);
          if (tmp < 0) {
            if (errno == EAGAIN || errno == EINTR)
              break; // no data available? signal?
            else {
              perror("read");
              return NULL;
            }
          } else if (tmp == 0) {
            if (sz == MAX_LINE_LEN)
              // out of buffer space!
              Error() << "Out of buffer space reading line!\n";
            else
              Debug() << "Client closed connection...\n";
            return NULL;
          }
          // at this point we know we got data, resize the string
          newpos = line + sz;
          sz += tmp;

          // try and grab the line if we have it now..
          while ( (newlinePos = strchr(newpos, '\n')) ) {
            std::string theLine;
            *newlinePos = 0;
            theLine = line; // assign it to theLine
            memmove(line, newlinePos+1, sz - (newlinePos-line+1));
            sz -= newlinePos-line+1;
            newpos = line;
            line[sz] = 0;

            connThreadLineProc(s, theLine);

          }
          
          // we dont' have the line, update newpos to avoid a redundant 
          // string search above..
          newpos = line + sz;
        }
      }
    }

    pthread_cleanup_pop(1);

    return NULL;
}

static  CID parseCID(int sock, 
                     const std::string & line, 
                     unsigned * arg2 = 0,
                     unsigned * arg3 = 0, 
                     unsigned * arg4 = 0,
                     unsigned * arg5 = 0);

static void validateCID(int sock, const CID & c)
{
      if ( devices.find(c.dev) == devices.end() 
           || c.subdev >= devices[c.dev].nSubdevs
           || c.chan >= devices[c.dev].subdev[c.subdev].nChans ) {
        // invalid CID?!? wtf?
        Error() << "Invalid CID!\n";     
        Protocol::SendError(sock, "CID is not valid");
        pthread_exit(0);
      }
}

static void expireOldDataThreadKeys();

static inline unsigned long long toULL(const struct timespec &tp)
{
  return static_cast<unsigned long long>(tp.tv_sec)*1000000000ULL 
    + static_cast<unsigned long long>(tp.tv_nsec);
}

static void connThreadLineProc(int sock, std::string & line)
{
  std::transform(line.begin(), line.end(), line.begin(), tolower);
  std::string::size_type pos = line.find_first_not_of(" \t\b\r", 0);

  Debug() << "Got Line: " << line << "\n";

  if (pos == line.npos) pos = 0; // for line that is all whitespace.....

  if (line.find(Protocol::GetComediDevices, pos) == pos) { 
    // get comedi devices

      int tmp = devices.size();
      dw_to_le(&tmp); // convert machine int to protocol format int (LE)
      tmp = ::send(sock, &tmp, sizeof(tmp), MSG_NOSIGNAL);
      if (tmp != sizeof(tmp)) {
        // send error?
        Error() << "Send error when sending comedi devices size!\n";
        pthread_exit(0);
      }
      ComediDevice *devs = new ComediDevice[devices.size()];
      ComediDeviceMap::iterator it;
      ComediDevice * cur;
      for (cur = devs, it = devices.begin(); it != devices.end(); ++it) { 
        // copy the devices to a straight array array...
          ::memcpy(cur++, &it->second, sizeof(it->second));
      }
      comediDeviceToLE(devs, devices.size()); // make sure it's in LE format in case our arch is big endian.. other side is expecting little-endian format..
      tmp = devices.size() * sizeof(*devs);
      tmp = ::send(sock, devs, tmp, MSG_NOSIGNAL);
      if (tmp != static_cast<int>(devices.size() * sizeof(*devs))) {
        // send error?
        delete [] devs;
        Error() << "Send error when sending comedi devices!\n";
        pthread_exit(0);
      }
      delete [] devs;    

  } else if (line.find(Protocol::YouAreADataThread, pos) == pos) {
      // we are a data thread! yay!

      // grep key:
      if ( (pos = line.find_first_of("-0123456789")) == line.npos) {
        // no key?!? wtf?
        Error() << "Parse error when grepping for datathread key!\n";     
        Protocol::SendError(sock, "Could not parse datathread key");
        pthread_exit(0);      
      }
      std::string num = line.substr(pos);
      std::stringstream ss(num);
      unsigned long long id = 0;
      ss >> id;
      pthread_mutex_lock(&mut);

      if (dataThreadKeys.find(id) == dataThreadKeys.end()) {
        pthread_mutex_unlock(&mut);
        Protocol::SendError(sock, "Key invalid.  Have a nice day!");
        pthread_exit(0);
      }      
      ThreadArg * arg = threads[pthread_self()] = dataThreadKeys[id];      
      dataThreadKeys.erase(id);
      pthread_mutex_unlock(&mut);

      Protocol::SendOK(sock);

      arg->fd = sock;

      // now, just replace ourselves with the acquisition thread..
      pthread_exit( AcquisitionThread(arg) );

  } else if (line.find(Protocol::StartACQ, pos) == pos) {
    // they want to start an acquisition..

      unsigned range = 0, aref = 0, hz = 1;
      unsigned chanmask = 0;
      CID c = parseCID(sock, line, &range, &aref, &hz, &chanmask);
      if (range >= devices[c.dev].subdev[c.subdev].nRanges) range = 0;

      // since the acquisition will take place on a separate connection,
      // tell them about this connection by giving them a key to connect
      // with..
      ThreadArg * arg = new ThreadArg;
      unsigned type = devices[c.dev].subdev[c.subdev].type;
      if (type == COMEDI_SUBD_DIO || type == COMEDI_SUBD_DI || type == COMEDI_SUBD_DO)
        arg->chan = chanmask ? chanmask : ( 0x1 << (unsigned)c.chan );
      else
        arg->chan = c.chan;
      arg->range = range;
      arg->aref = aref;
      arg->sdev = &devices[c.dev].subdev[c.subdev];
      arg->hz = hz ? hz : 100;

      // do config for wanted channels..
//       if ( (arg->sdev->type == COMEDI_SUBD_DI 
//             || arg->sdev->type == COMEDI_SUBD_DIO) ) {
//         unsigned mask = chanmask ? chanmask : 0x1 << arg->chan;
//         for (unsigned i = 0; i < arg->sdev->nChans; ++i)
//           if ( 0x1<<i & mask ) {
//             comedi_t *handle = (comedi_t *)handles[arg->sdev->minor];    
//             unsigned sdev = arg->sdev->id;
//             comedi_dio_config(handle, sdev, i, COMEDI_INPUT);
//             dioConfigs[blah] = COMEDI_INPUT;
//           }
//       }

      struct timespec tp;
      clock_gettime(CLOCK_REALTIME, &tp);
      unsigned long long id = toULL(tp);
      pthread_mutex_lock(&mut);
      while( dataThreadKeys.find(id) != dataThreadKeys.end() )
        ++id;
      dataThreadKeys[id] = arg;
      pthread_mutex_unlock(&mut);
      std::stringstream ss;
      ss << id << "\n";
      std::string num = ss.str();
      ::send(sock, num.c_str(), num.size(), MSG_NOSIGNAL); // send them the id-key
      expireOldDataThreadKeys();
  } else if (line.find(Protocol::WriteData, pos) == pos) {

      unsigned data = 0, range = 0, aref = 0;
      unsigned chanMask = 0;
      CID c = parseCID(sock, line, &data, &range, &aref, &chanMask);
      if (range >= devices[c.dev].subdev[c.subdev].nRanges) range = 0;
      unsigned type = devices[c.dev].subdev[c.subdev].type;
      comedi_t *handle = (comedi_t *)handles[c.dev];
      if ((type == COMEDI_SUBD_DIO || type == COMEDI_SUBD_DO) && chanMask) {
        // it's a bitfield array
        /*for (unsigned i = 0; i < devices[c.dev].subdev[c.subdev].nChans; ++i)
          // do config
          if (chanMask & 0x1<<i ) comedi_dio_config(handle, c.subdev, i, COMEDI_OUTPUT);*/
        // do bitfield
        comedi_dio_bitfield(handle, c.subdev, chanMask, &data);
      } else {        
        // it's a single channel
        /*if (type == COMEDI_SUBD_DIO) // dio, make sure we are configured for OUTPUT
          comedi_dio_config(handle, c.subdev, c.chan, COMEDI_OUTPUT);*/
        comedi_data_write(handle, c.subdev, c.chan, range, aref, data);
      }
      Protocol::SendOK(sock);

  } else if (line.find(Protocol::ReadData, pos) == pos) {

      unsigned range = 0, aref = 0, chanMask = 0;
      lsampl_t data = 0;
      CID c = parseCID(sock, line, &range, &aref, &chanMask);
      unsigned type = devices[c.dev].subdev[c.subdev].type;
      comedi_t *handle = (comedi_t *)handles[c.dev];
      if (range >= devices[c.dev].subdev[c.subdev].nRanges) range = 0;
      if ((type == COMEDI_SUBD_DIO || type == COMEDI_SUBD_DO) && chanMask) {
        // it's a bitfield array
//         for (unsigned i = 0; i < devices[c.dev].subdev[c.subdev].nChans; ++i)
//           // do config
//           if (chanMask & 0x1<<i) comedi_dio_config(handle, c.subdev, i, COMEDI_INPUT);
        // do bitfield
        comedi_dio_bitfield(handle, c.subdev, 0, &data);
        data &= chanMask; // only take bits they wanted
      } else {        
        // it's a single channel
//         if (type == COMEDI_SUBD_DIO) // dio, make sure we are configured for OUTPUT
//           comedi_dio_config(handle, c.subdev, c.chan, COMEDI_INPUT);
        comedi_data_read(handle, c.subdev, c.chan, range, aref, &data);
      }
      std::stringstream ss;
      ss << data << "\n";
      std::string num = ss.str();
      ::send(sock, num.c_str(), num.size(), MSG_NOSIGNAL); // send them the result      
      //Protocol::SendOK(sock);
  } else if (line.find(Protocol::DIOConfig, pos) == pos) {
    unsigned mode = 0xffffffff, mask = 0;
    CID c = parseCID(sock, line, &mode, &mask);
    comedi_t *handle = (comedi_t *)handles[c.dev];
    unsigned type = devices[c.dev].subdev[c.subdev].type;
    if (!mask) mask = 0x1<<c.chan;
    bool didOne = false;

    if (!handle || type != COMEDI_SUBD_DIO 
        || (mode != ComediSubDevice::Input && mode != ComediSubDevice::Output))
    {
      Protocol::SendError(sock, "Invalid parameter.");
      pthread_exit(0);
    }
    
    for (unsigned chan = 0; chan < devices[c.dev].subdev[c.subdev].nChans; ++chan) {      
      if ( (mask & (0x1<<chan)) && comedi_dio_config(handle, c.subdev, chan, mode) > -1) {
        dioConfigs[c] = static_cast<ComediSubDevice::DIOConfig>(mode);
        didOne = true;
      }
    }
    if (!didOne) {
      Protocol::SendError(sock, "Invalid parameter or error setting DIO mode.");
      pthread_exit(0);
    } else {
      Protocol::SendOK(sock);
    }
  } else if (line.find(Protocol::GetDIOConfig, pos) == pos) {
    unsigned mask = 0;
    CID c = parseCID(sock, line, &mask);
    comedi_t *handle = (comedi_t *)handles[c.dev];
    unsigned type = devices[c.dev].subdev[c.subdev].type;
    if (!mask) mask = 0x1<<c.chan;
    bool didOne = false;

    if (!handle || type != COMEDI_SUBD_DIO)
    {
      Protocol::SendError(sock, "Invalid parameter.");
      pthread_exit(0);
    }

    std::ostringstream os;

    for (unsigned chan = 0; chan < devices[c.dev].subdev[c.subdev].nChans; ++chan) {      
      if ( mask & (0x1<<chan) ) {
        c.chan = chan;
        int m = dioConfigs[c];
        didOne = true;
        os << m << " ";
      }
    }
    os << "\n";
    if (!didOne) {
      Protocol::SendError(sock, "Invalid parameter.");
      pthread_exit(0);
    } else {
      ::send(sock, os.str().c_str(), os.str().length(), MSG_NOSIGNAL);
    }
  } else if (line.find(Protocol::GetVersion, pos) == pos) {
    std::ostringstream out;
    out << Protocol::Version << "\n";
    ::send(sock, out.str().c_str(), out.str().size(), MSG_NOSIGNAL); // send them the result
  } else if (line.find(Protocol::Noop, pos) == pos) {
    std::ostringstream out;
    Protocol::SendOK(sock);
  } else {
    Error() << "Parse error for line\n";
    Protocol::SendError(sock, "Invalid protocol command.");
    pthread_exit(0);
  }
}

static void expireOldDataThreadKeys()
{
  struct timespec tp;
  clock_gettime(CLOCK_REALTIME, &tp);
  unsigned long long id = toULL(tp);
  id -= 5*1000000000; // expire stuff 5 seconds old or older
  pthread_mutex_lock(&mut);
  DataThreadKeys::iterator oldend = dataThreadKeys.upper_bound(id), it;
  for (it = dataThreadKeys.begin(); it != dataThreadKeys.end() && it != oldend; ++it) 
    delete it->second; // delete old ThreadArgs  
  dataThreadKeys.erase(dataThreadKeys.begin(), oldend); // clean out old/expired/never-used data thread keys..
  pthread_mutex_unlock(&mut);
}

static CID parseCID(int sock, const std::string & line, unsigned *arg2,
                    unsigned *arg3, unsigned *arg4, unsigned *arg5)
{
  std::string::size_type pos;

      // parse CID:
  if ( (pos = line.find_first_of("-0123456789")) == line.npos) {
        // no CID?!? wtf?
        Error() << "Parse error when grepping for CID!\n";
        Protocol::SendError(sock, "Could not parse CID");
        pthread_exit(0);
  }
  std::string num = line.substr(pos);
  std::stringstream ss(num);
  unsigned short cid = 0;
  ss >> cid;
  CID c(cid);
  validateCID(sock, c);
  if (arg2) ss >> *arg2;
  if (arg3) ss >> *arg3;
  if (arg4) ss >> *arg4;
  if (arg5) ss >> *arg5;
  return c;
}

static void doConnection(int connSock)
{
  pthread_t thr;
  if ( pthread_create(&thr, 0, connThread, (void *)connSock) ) {
    Error() << "Error creating connection thread!\n";
    ::exit(1);
  }
  pthread_mutex_lock(&mut);
  threads[thr] = NULL;
  pthread_mutex_unlock(&mut);
}

static void dioConfigAll(ComediSubDevice::DIOConfig conf)
{
  CID cid;
  for (ComediDeviceMap::iterator it = devices.begin(); it != devices.end(); ++it) {
    unsigned id = 0;
    int ret = 0;
    while (ret > 0) {
      unsigned minor = it->first;
      ComediDevice & dev = it->second;
      comedi_t *handle = (comedi_t *)handles[minor];
      cid.dev = minor;
      if ( (ret = comedi_find_subdevice_by_type(handle, COMEDI_SUBD_DIO, id)) >= 0 ) {
        id = ret;
        cid.subdev = id;
        for (unsigned i = 0; i < dev.subdev[id].nChans; ++i) {
          cid.chan = i;
          comedi_dio_config(handle, id, i, (unsigned)conf);
          dioConfigs[cid] = conf; // cache it
        }
      }
    }
  }
}

void ParseArgs(int & argc, char ** & argv)
{
  int c;
  while ( (c = getopt(argc, argv, "d")) > -1 ) {
    switch(c) {
    case 'd': ::debug = true;
      break;
    }
  }
  char *orig_argv0 = argv[0];
  argv = &argv[optind-1];
  argc -= optind-1;
  argv[0] = orig_argv0;
}

int main(int argc, char *argv[])
{
   ParseArgs(argc, argv);
   ProbeDevices(devices, handles);

   if (devices.size() == 0) {
     Error() << "No comedi devices found\n";
     return 1;
   }

   Log() << "Found " << devices.size() << " comedi devices.\n";

   serverSocket = BindSocket(argc, argv);

   ::atexit(ReleaseSocket);

   ::signal(SIGINT, sighandler);
   ::signal(SIGQUIT, sighandler);
   ::signal(SIGTERM, sighandler);

   dioConfigAll(ComediSubDevice::Output); // default everyone to output

   if (::listen(serverSocket, 1)) {
     ::perror("listen");
     return 1;
   }
   
   // for now we only handle one connection at a time
   while (!stopItAll) {

     struct pollfd pfd = { fd: serverSocket, events: POLLIN|POLLPRI, 
                           revents: 0 };
     
     int tmp = ::poll(&pfd, 1, 1000); // wakeup every 1000 ms..
     if (tmp < 0) { if (errno == EINTR) continue; else stopItAll = true; }
     else if (tmp > 0) {
       int connSock = ::accept(serverSocket, NULL, NULL); // we don't care about the originating address for now
       if (connSock < 0) {
         ::perror("accept");
         return 1;
       }
       
       doConnection(connSock);
     }

   }

   ::close(serverSocket);
   serverSocket = -1;
   
   pthread_mutex_lock(&mut);
   ThreadMap thrcpy = threads;
   pthread_mutex_unlock(&mut);
   for (ThreadMap::iterator it = thrcpy.begin(); it != thrcpy.end(); ++it) {
     pthread_cancel(it->first);
     pthread_join(it->first, NULL);
   }

   // not reached
   return 0;
}
